#!/usr/bin/env python
# -*- coding: utf-8 -*-

#!/usr/bin/env python
# -*- coding: utf-8 -*-

# NER数据集持久化，与标注文件之间的互转。

import os, json, re
from dataclasses import dataclass, field
from typing import List

from loguru import logger
from tqdm import tqdm



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import torch


def common_batch_encode(texts, label2id, tokenizer, max_seq_length):
    all_encodes = tokenizer.batch_encode(texts, add_special_tokens=True)
    all_tokens = all_encodes['tokens']
    all_token_offsets = all_encodes['offsets']
    all_token2char = all_encodes['token2char']
    all_char2token = all_encodes['char2token']

    all_input_ids = all_encodes['ids']
    all_attention_mask = all_encodes['attention_mask']
    all_token_type_ids = all_encodes['type_ids']

    all_input_lens = [len(tokens) for tokens in all_tokens]

    all_padding_lens = [max_seq_length - n for n in all_input_lens]
    for i, (input_ids, attention_mask, token_type_ids, token2char,
            token_offsets, padding_length) in enumerate(
                tqdm(zip(all_input_ids, all_attention_mask, all_token_type_ids,
                         all_token2char, all_token_offsets, all_padding_lens),
                     desc="common_batch_encode")):

        if padding_length < 0:
            all_tokens[i] = all_tokens[i][:max_seq_length - 1] + ['[SEP]']
            all_input_ids[i] = input_ids[:max_seq_length - 1] + [102]
            all_attention_mask[i] = attention_mask[:max_seq_length]
            all_token_type_ids[i] = token_type_ids[:max_seq_length]
            all_token2char[i] = token2char[:max_seq_length]
            all_token_offsets[i] = token_offsets[:max_seq_length - 1] + [(0, 0)
                                                                         ]
            #  logger.warning(f"all_tokens[i]: {all_tokens[i]}")
            #  logger.debug(f"all_input_ids[i]: {all_input_ids[i]}")
            #  logger.debug(f"all_token_offsets[i]: {all_token_offsets[i]}")

        else:
            all_input_ids[i] = input_ids + [0] * padding_length
            all_attention_mask[i] = attention_mask + [0] * padding_length
            all_token_type_ids[i] = token_type_ids + [0] * padding_length
            all_token2char[i] = token2char + [0] * padding_length
            all_token_offsets[i] = token_offsets + [(0, 0)] * padding_length
        all_input_lens[i] = len(input_ids)

    #  logger.warning(f"all_input_ids: {all_input_ids}")
    all_input_ids = torch.from_numpy(np.array(all_input_ids, dtype=np.int64))
    all_attention_mask = torch.from_numpy(
        np.array(all_attention_mask, dtype=np.int64))
    all_token_type_ids = torch.from_numpy(
        np.array(all_token_type_ids, dtype=np.int64))
    all_input_lens = torch.from_numpy(np.array(all_input_lens, dtype=np.int64))
    all_token_offsets = torch.from_numpy(
        np.array(all_token_offsets, dtype=np.int64))

    return (all_tokens, all_token2char, all_char2token, all_input_ids,
            all_attention_mask, all_token_type_ids, all_input_lens,
            all_token_offsets)


def common_to_tensors(all_input_ids, all_attention_mask, all_token_type_ids,
                      all_input_lens, all_token_offsets):
    all_input_ids = torch.from_numpy(np.array(all_input_ids, dtype=np.int64))
    all_attention_mask = torch.from_numpy(
        np.array(all_attention_mask, dtype=np.int64))
    all_token_type_ids = torch.from_numpy(
        np.array(all_token_type_ids, dtype=np.int64))
    all_input_lens = torch.from_numpy(np.array(all_input_lens, dtype=np.int64))
    all_token_offsets = torch.from_numpy(
        np.array(all_token_offsets, dtype=np.int64))

    return all_input_ids, all_attention_mask, all_token_type_ids, all_input_lens, all_token_offsets


def encode_examples(examples, label2id, tokenizer, max_seq_length):

    num_labels = len(label2id)
    texts = [e.text_a[:max_seq_length - 2] for e in examples]

    all_tokens, all_token2char, all_char2token, all_input_ids, all_attention_mask, all_token_type_ids, all_input_lens, all_token_offsets = common_batch_encode(
        texts, label2id, tokenizer, max_seq_length)

    #  all_labels = [label2id[e.label] if e.label else 0 for e in examples]
    all_labels = []
    for e in examples:
        if isinstance(e.label, list):
            targets = [0] * len(label2id)
            for x in e.label:
                targets[label2id[x]] = 1
            all_labels.append(targets)
        else:
            if e.label:
                all_labels.append(label2id[e.label])
            else:
                all_labels.append(0)

    all_input_ids, all_attention_mask, all_token_type_ids, all_input_lens, all_token_offsets = common_to_tensors(
        all_input_ids, all_attention_mask, all_token_type_ids, all_input_lens,
        all_token_offsets)

    all_labels = torch.from_numpy(np.array(all_labels, dtype=np.int64))

    all_features = [
        InputFeature(input_ids=input_ids,
                     attention_mask=attention_mask,
                     token_type_ids=token_type_ids,
                     input_len=input_len,
                     token_offsets=token_offsets,
                     label=label,
                     text=text) for input_ids, attention_mask,
        token_type_ids, input_len, token_offsets, label, text in zip(
            all_input_ids, all_attention_mask, all_token_type_ids,
            all_input_lens, all_token_offsets, all_labels, texts)
    ]

    return all_features



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import torch
import torch.nn as nn
from torch.nn import CrossEntropyLoss, MSELoss
from transformers.models.bert import BertModel, BertPreTrainedModel, BertForMaskedLM
from transformers.modeling_outputs import SequenceClassifierOutput


class BertForSequenceClassification(BertPreTrainedModel):
    def __init__(self, config):
        super().__init__(config)
        self.num_labels = config.num_labels

        self.bert = BertModel(config)
        self.dropout = nn.Dropout(config.hidden_dropout_prob)
        self.classifier = nn.Linear(config.hidden_size, config.num_labels)

        self.init_weights()

    def forward(
        self,
        input_ids=None,
        attention_mask=None,
        token_type_ids=None,
        position_ids=None,
        head_mask=None,
        inputs_embeds=None,
        labels=None,
        output_attentions=None,
        output_hidden_states=None,
        return_dict=None,
    ):
        return_dict = return_dict if return_dict is not None else self.config.use_return_dict

        outputs = self.bert(
            input_ids,
            attention_mask=attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        pooled_output = outputs[1]

        pooled_output = self.dropout(pooled_output)
        logits = self.classifier(pooled_output)

        loss = None
        if labels is not None:
            if self.num_labels == 1:
                #  We are doing regression
                loss_fct = MSELoss()
                loss = loss_fct(logits.view(-1), labels.view(-1))
            else:
                loss_fct = CrossEntropyLoss()
                loss = loss_fct(logits.view(-1, self.num_labels),
                                labels.view(-1))

        if not return_dict:
            output = (logits, ) + outputs[2:]
            return ((loss, ) + output) if loss is not None else output

        return SequenceClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )



# -*- coding: utf-8 -*-
# File: serialize.py

import numpy as np
import os
import platform
from collections import defaultdict

from ..utils import logger
from ..utils.serialize import dumps, loads
from ..utils.develop import create_dummy_class  # noqa
from ..utils.utils import get_tqdm
from .base import DataFlow
from .common import FixedSizeData, MapData
from .format import HDF5Data, LMDBData
from .raw import DataFromGenerator, DataFromList

__all__ = ['LMDBSerializer', 'NumpySerializer', 'TFRecordSerializer', 'HDF5Serializer']


def _reset_df_and_get_size(df):
    df.reset_state()
    try:
        sz = len(df)
    except NotImplementedError:
        sz = 0
    return sz


class LMDBSerializer():
    """
    Serialize a Dataflow to a lmdb database, where the keys are indices and values
    are serialized datapoints.

    You will need to ``pip install lmdb`` to use it.

    Example:

    .. code-block:: python

        LMDBSerializer.save(my_df, "output.lmdb")

        new_df = LMDBSerializer.load("output.lmdb", shuffle=True)
    """
    @staticmethod
    def save(df, path, write_frequency=5000):
        """
        Args:
            df (DataFlow): the DataFlow to serialize.
            path (str): output path. Either a directory or an lmdb file.
            write_frequency (int): the frequency to write back data to disk.
                A smaller value reduces memory usage.
        """
        assert isinstance(df, DataFlow), type(df)
        isdir = os.path.isdir(path)
        if isdir:
            assert not os.path.isfile(os.path.join(path, 'data.mdb')), "LMDB file exists!"
        else:
            assert not os.path.isfile(path), "LMDB file {} exists!".format(path)
        # It's OK to use super large map_size on Linux, but not on other platforms
        # See: https://github.com/NVIDIA/DIGITS/issues/206
        map_size = 1099511627776 * 2 if platform.system() == 'Linux' else 128 * 10**6
        db = lmdb.open(path, subdir=isdir,
                       map_size=map_size, readonly=False,
                       meminit=False, map_async=True)    # need sync() at the end
        size = _reset_df_and_get_size(df)

        # put data into lmdb, and doubling the size if full.
        # Ref: https://github.com/NVIDIA/DIGITS/pull/209/files
        def put_or_grow(txn, key, value):
            try:
                txn.put(key, value)
                return txn
            except lmdb.MapFullError:
                pass
            txn.abort()
            curr_size = db.info()['map_size']
            new_size = curr_size * 2
            logger.info("Doubling LMDB map_size to {:.2f}GB".format(new_size / 10**9))
            db.set_mapsize(new_size)
            txn = db.begin(write=True)
            txn = put_or_grow(txn, key, value)
            return txn

        with get_tqdm(total=size) as pbar:
            idx = -1

            # LMDB transaction is not exception-safe!
            # although it has a context manager interface
            txn = db.begin(write=True)
            for idx, dp in enumerate(df):
                txn = put_or_grow(txn, u'{:08}'.format(idx).encode('ascii'), dumps(dp))
                pbar.update()
                if (idx + 1) % write_frequency == 0:
                    txn.commit()
                    txn = db.begin(write=True)
            txn.commit()

            keys = [u'{:08}'.format(k).encode('ascii') for k in range(idx + 1)]
            with db.begin(write=True) as txn:
                txn = put_or_grow(txn, b'__keys__', dumps(keys))

            logger.info("Flushing database ...")
            db.sync()
        db.close()

    @staticmethod
    def load(path, shuffle=True):
        """
        Note:
            If you found deserialization being the bottleneck, you can use :class:`LMDBData` as the reader
            and run deserialization as a mapper in parallel.
        """
        df = LMDBData(path, shuffle=shuffle)
        return MapData(df, LMDBSerializer._deserialize_lmdb)

    @staticmethod
    def _deserialize_lmdb(dp):
        return loads(dp[1])


class NumpySerializer():
    """
    Serialize the entire dataflow to a npz dict.
    Note that this would have to store the entire dataflow in memory,
    and is also >10x slower than LMDB/TFRecord serializers.
    """

    @staticmethod
    def save(df, path):
        """
        Args:
            df (DataFlow): the DataFlow to serialize.
            path (str): output npz file.
        """
        buffer = []
        size = _reset_df_and_get_size(df)
        with get_tqdm(total=size) as pbar:
            for dp in df:
                buffer.append(dp)
                pbar.update()
        np.savez_compressed(path, buffer=np.asarray(buffer, dtype=np.object))

    @staticmethod
    def load(path, shuffle=True):
        # allow_pickle defaults to False since numpy 1.16.3
        # (https://www.numpy.org/devdocs/release.html#unpickling-while-loading-requires-explicit-opt-in)
        buffer = np.load(path, allow_pickle=True)['buffer']
        return DataFromList(buffer, shuffle=shuffle)


class TFRecordSerializer():
    """
    Serialize datapoints to bytes (by tensorpack's default serializer) and write to a TFRecord file.

    Note that TFRecord does not support random access and is in fact not very performant.
    It's better to use :class:`LMDBSerializer`.
    """
    @staticmethod
    def save(df, path):
        """
        Args:
            df (DataFlow): the DataFlow to serialize.
            path (str): output tfrecord file.
        """
        size = _reset_df_and_get_size(df)
        with tf.python_io.TFRecordWriter(path) as writer, get_tqdm(total=size) as pbar:
            for dp in df:
                writer.write(dumps(dp))
                pbar.update()

    @staticmethod
    def load(path, size=None):
        """
        Args:
            size (int): total number of records. If not provided, the returned dataflow will have no `__len__()`.
                It's needed because this metadata is not stored in the TFRecord file.
        """
        gen = tf.python_io.tf_record_iterator(path)
        ds = DataFromGenerator(gen)
        ds = MapData(ds, loads)
        if size is not None:
            ds = FixedSizeData(ds, size)
        return ds


class HDF5Serializer():
    """
    Write datapoints to a HDF5 file.

    Note that HDF5 files are in fact not very performant and currently do not support lazy loading.
    It's better to use :class:`LMDBSerializer`.
    """
    @staticmethod
    def save(df, path, data_paths):
        """
        Args:
            df (DataFlow): the DataFlow to serialize.
            path (str): output hdf5 file.
            data_paths (list[str]): list of h5 paths. It should have the same
                length as each datapoint, and each path should correspond to one
                component of the datapoint.
        """
        size = _reset_df_and_get_size(df)
        buffer = defaultdict(list)

        with get_tqdm(total=size) as pbar:
            for dp in df:
                assert len(dp) == len(data_paths), "Datapoint has {} components!".format(len(dp))
                for k, el in zip(data_paths, dp):
                    buffer[k].append(el)
                pbar.update()

        with h5py.File(path, 'w') as hf, get_tqdm(total=len(data_paths)) as pbar:
            for data_path in data_paths:
                hf.create_dataset(data_path, data=buffer[data_path])
                pbar.update()

    @staticmethod
    def load(path, data_paths, shuffle=True):
        """
        Args:
            data_paths (list): list of h5 paths to be zipped.
        """
        return HDF5Data(path, data_paths, shuffle)


try:
    import lmdb
except ImportError:
    LMDBSerializer = create_dummy_class('LMDBSerializer', 'lmdb')   # noqa

try:
    import tensorflow as tf
except ImportError:
    TFRecordSerializer = create_dummy_class('TFRecordSerializer', 'tensorflow')   # noqa

try:
    import h5py
except ImportError:
    HDF5Serializer = create_dummy_class('HDF5Serializer', 'h5py')   # noqa


if __name__ == '__main__':
    from .raw import FakeData
    import time
    ds = FakeData([[300, 300, 3], [1]], 1000)

    print(time.time())
    TFRecordSerializer.save(ds, 'out.tfrecords')
    print(time.time())
    df = TFRecordSerializer.load('out.tfrecords', size=1000)
    df.reset_state()
    for idx, dp in enumerate(df):
        pass
    print("TF Finished, ", idx)
    print(time.time())

    LMDBSerializer.save(ds, 'out.lmdb')
    print(time.time())
    df = LMDBSerializer.load('out.lmdb')
    df.reset_state()
    for idx, dp in enumerate(df):
        pass
    print("LMDB Finished, ", idx)
    print(time.time())

    NumpySerializer.save(ds, 'out.npz')
    print(time.time())
    df = NumpySerializer.load('out.npz')
    df.reset_state()
    for idx, dp in enumerate(df):
        pass
    print("Numpy Finished, ", idx)
    print(time.time())

    paths = ['p1', 'p2']
    HDF5Serializer.save(ds, 'out.h5', paths)
    print(time.time())
    df = HDF5Serializer.load('out.h5', paths)
    df.reset_state()
    for idx, dp in enumerate(df):
        pass
    print("HDF5 Finished, ", idx)
    print(time.time())



# -*- coding: utf-8 -*-
# File: geometry.py


import math
import numpy as np
import cv2

from .base import ImageAugmentor
from .transform import WarpAffineTransform, CropTransform, TransformList

__all__ = ['Shift', 'Rotation', 'RotationAndCropValid', 'Affine']


class Shift(ImageAugmentor):
    """ Random horizontal and vertical shifts """

    def __init__(self, horiz_frac=0, vert_frac=0,
                 border=cv2.BORDER_REPLICATE, border_value=0):
        """
        Args:
            horiz_frac (float): max abs fraction for horizontal shift
            vert_frac (float): max abs fraction for horizontal shift
            border: cv2 border method
            border_value: cv2 border value for border=cv2.BORDER_CONSTANT
        """
        assert horiz_frac < 1.0 and vert_frac < 1.0
        super(Shift, self).__init__()
        self._init(locals())

    def get_transform(self, img):
        max_dx = self.horiz_frac * img.shape[1]
        max_dy = self.vert_frac * img.shape[0]
        dx = np.round(self._rand_range(-max_dx, max_dx))
        dy = np.round(self._rand_range(-max_dy, max_dy))

        mat = np.array([[1, 0, dx], [0, 1, dy]], dtype='float32')
        return WarpAffineTransform(
            mat, img.shape[1::-1],
            borderMode=self.border, borderValue=self.border_value)


class Rotation(ImageAugmentor):
    """ Random rotate the image w.r.t a random center"""

    def __init__(self, max_deg, center_range=(0, 1),
                 interp=cv2.INTER_LINEAR,
                 border=cv2.BORDER_REPLICATE, step_deg=None, border_value=0):
        """
        Args:
            max_deg (float): max abs value of the rotation angle (in degree).
            center_range (tuple): (min, max) range of the random rotation center.
            interp: cv2 interpolation method
            border: cv2 border method
            step_deg (float): if not None, the stepping of the rotation
                angle. The rotation angle will be a multiple of step_deg. This
                option requires ``max_deg==180`` and step_deg has to be a divisor of 180)
            border_value: cv2 border value for border=cv2.BORDER_CONSTANT
        """
        assert step_deg is None or (max_deg == 180 and max_deg % step_deg == 0)
        super(Rotation, self).__init__()
        self._init(locals())

    def get_transform(self, img):
        center = img.shape[1::-1] * self._rand_range(
            self.center_range[0], self.center_range[1], (2,))
        deg = self._rand_range(-self.max_deg, self.max_deg)
        if self.step_deg:
            deg = deg // self.step_deg * self.step_deg
        """
        The correct center is shape*0.5-0.5. This can be verified by:

        SHAPE = 7
        arr = np.random.rand(SHAPE, SHAPE)
        orig = arr
        c = SHAPE * 0.5 - 0.5
        c = (c, c)
        for k in range(4):
            mat = cv2.getRotationMatrix2D(c, 90, 1)
            arr = cv2.warpAffine(arr, mat, arr.shape)
        assert np.all(arr == orig)
        """
        mat = cv2.getRotationMatrix2D(tuple(center - 0.5), float(deg), 1)
        return WarpAffineTransform(
            mat, img.shape[1::-1], interp=self.interp,
            borderMode=self.border, borderValue=self.border_value)


class RotationAndCropValid(ImageAugmentor):
    """ Random rotate and then crop the largest possible rectangle.
        Note that this will produce images of different shapes.
    """

    def __init__(self, max_deg, interp=cv2.INTER_LINEAR, step_deg=None):
        """
        Args:
            max_deg, interp, step_deg: same as :class:`Rotation`
        """
        assert step_deg is None or (max_deg == 180 and max_deg % step_deg == 0)
        super(RotationAndCropValid, self).__init__()
        self._init(locals())

    def _get_deg(self, img):
        deg = self._rand_range(-self.max_deg, self.max_deg)
        if self.step_deg:
            deg = deg // self.step_deg * self.step_deg
        return float(deg)

    def get_transform(self, img):
        deg = self._get_deg(img)

        h, w = img.shape[:2]
        center = (img.shape[1] * 0.5, img.shape[0] * 0.5)
        rot_m = cv2.getRotationMatrix2D((center[0] - 0.5, center[1] - 0.5), deg, 1)
        tfm = WarpAffineTransform(rot_m, (w, h), interp=self.interp)

        neww, newh = RotationAndCropValid.largest_rotated_rect(w, h, deg)
        neww = min(neww, w)
        newh = min(newh, h)
        newx = int(center[0] - neww * 0.5)
        newy = int(center[1] - newh * 0.5)
        tfm2 = CropTransform(newy, newx, newh, neww)
        return TransformList([tfm, tfm2])

    @staticmethod
    def largest_rotated_rect(w, h, angle):
        """
        Get largest rectangle after rotation.
        http://stackoverflow.com/questions/16702966/rotate-image-and-crop-out-black-borders
        """
        angle = angle / 180.0 * math.pi
        if w <= 0 or h <= 0:
            return 0, 0

        width_is_longer = w >= h
        side_long, side_short = (w, h) if width_is_longer else (h, w)

        # since the solutions for angle, -angle and 180-angle are all the same,
        # if suffices to look at the first quadrant and the absolute values of sin,cos:
        sin_a, cos_a = abs(math.sin(angle)), abs(math.cos(angle))
        if side_short <= 2. * sin_a * cos_a * side_long:
            # half constrained case: two crop corners touch the longer side,
            #   the other two corners are on the mid-line parallel to the longer line
            x = 0.5 * side_short
            wr, hr = (x / sin_a, x / cos_a) if width_is_longer else (x / cos_a, x / sin_a)
        else:
            # fully constrained case: crop touches all 4 sides
            cos_2a = cos_a * cos_a - sin_a * sin_a
            wr, hr = (w * cos_a - h * sin_a) / cos_2a, (h * cos_a - w * sin_a) / cos_2a
        return int(np.round(wr)), int(np.round(hr))


class Affine(ImageAugmentor):
    """
    Random affine transform of the image w.r.t to the image center.
    Transformations involve:

    - Translation ("move" image on the x-/y-axis)
    - Rotation
    - Scaling ("zoom" in/out)
    - Shear (move one side of the image, turning a square into a trapezoid)
    """

    def __init__(self, scale=None, translate_frac=None, rotate_max_deg=0.0, shear=0.0,
                 interp=cv2.INTER_LINEAR, border=cv2.BORDER_REPLICATE, border_value=0):
        """
        Args:
            scale (tuple of 2 floats): scaling factor interval, e.g (a, b), then scale is
                randomly sampled from the range a <= scale <= b. Will keep
                original scale by default.
            translate_frac (tuple of 2 floats): tuple of max abs fraction for horizontal
                and vertical translation. For example translate_frac=(a, b), then horizontal shift
                is randomly sampled in the range 0 < dx < img_width * a and vertical shift is
                randomly sampled in the range 0 < dy < img_height * b. Will
                not translate by default.
            shear (float): max abs shear value in degrees between 0 to 180
            interp: cv2 interpolation method
            border: cv2 border method
            border_value: cv2 border value for border=cv2.BORDER_CONSTANT
        """
        if scale is not None:
            assert isinstance(scale, tuple) and len(scale) == 2, \
                "Argument scale should be a tuple of two floats, e.g (a, b)"

        if translate_frac is not None:
            assert isinstance(translate_frac, tuple) and len(translate_frac) == 2, \
                "Argument translate_frac should be a tuple of two floats, e.g (a, b)"

        assert shear >= 0.0, "Argument shear should be between 0.0 and 180.0"

        super(Affine, self).__init__()
        self._init(locals())

    def get_transform(self, img):
        if self.scale is not None:
            scale = self._rand_range(self.scale[0], self.scale[1])
        else:
            scale = 1.0

        if self.translate_frac is not None:
            max_dx = self.translate_frac[0] * img.shape[1]
            max_dy = self.translate_frac[1] * img.shape[0]
            dx = np.round(self._rand_range(-max_dx, max_dx))
            dy = np.round(self._rand_range(-max_dy, max_dy))
        else:
            dx = 0
            dy = 0

        if self.shear > 0.0:
            shear = self._rand_range(-self.shear, self.shear)
            sin_shear = math.sin(math.radians(shear))
            cos_shear = math.cos(math.radians(shear))
        else:
            sin_shear = 0.0
            cos_shear = 1.0

        center = (img.shape[1::-1] * np.array((0.5, 0.5))) - 0.5
        deg = self._rand_range(-self.rotate_max_deg, self.rotate_max_deg)

        transform_matrix = cv2.getRotationMatrix2D(tuple(center), deg, scale)

        # Apply shear :
        if self.shear > 0.0:
            m00 = transform_matrix[0, 0]
            m01 = transform_matrix[0, 1]
            m10 = transform_matrix[1, 0]
            m11 = transform_matrix[1, 1]
            transform_matrix[0, 1] = m01 * cos_shear + m00 * sin_shear
            transform_matrix[1, 1] = m11 * cos_shear + m10 * sin_shear
            # Add correction term to keep the center unchanged
            tx = center[0] * (1.0 - m00) - center[1] * transform_matrix[0, 1]
            ty = -center[0] * m10 + center[1] * (1.0 - transform_matrix[1, 1])
            transform_matrix[0, 2] = tx
            transform_matrix[1, 2] = ty

        # Apply shift :
        transform_matrix[0, 2] += dx
        transform_matrix[1, 2] += dy
        return WarpAffineTransform(transform_matrix, img.shape[1::-1],
                                   self.interp, self.border, self.border_value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
from tqdm import tqdm, trange
from loguru import logger


def get_ner_results(metric):
    eval_info, entity_info = metric.result()
    results = {f'{key}': value for key, value in eval_info.items()}

    title = f"{' '*24}   acc    recall f1    "
    title += " Right/Pred/True "
    title_len = len(title.encode('gbk'))
    logger.info(f"{'=' * title_len}")
    logger.info(title)
    logger.info(f"{'-' * title_len}")

    sorted_entity_info = sorted(entity_info.items(),
                                key=lambda x: x[1]['f1'],
                                reverse=True)
    total_acc = 0
    total_recall = 0
    total_f1 = 0
    for key, metrics in sorted_entity_info:
        if ':' in key:
            category = key.split(':')[1]
        else:
            category = key
        if metric.ignore_categories and category in metric.ignore_categories:
            disp_key = ' -' + key[:16]
        else:
            disp_key = ' ' + key[:16]
        disp_key += ' ' * (24 - len(disp_key.encode('gbk')))
        info = f"{disp_key} | {metrics['acc']:.4f} {metrics['recall']:.4f} {metrics['f1']:.4f}"
        right = metrics['right']
        found = metrics['found']
        origin = metrics['origin']
        info += f" {right}/{found}/{origin}"
        logger.info(info)
        if metric.ignore_categories and category in metric.ignore_categories:
            pass
        else:
            total_acc += metrics['acc']
            total_recall += metrics['recall']
            total_f1 += metrics['f1']
    if metric.ignore_categories:
        num_categories = len(sorted_entity_info) - len(
            metric.ignore_categories)
    else:
        num_categories = len(sorted_entity_info)
    macro_acc = total_acc / num_categories
    macro_recall = total_recall / num_categories
    macro_f1 = total_f1 / num_categories
    logger.info(f"{'-' * title_len}")

    info = f" Micro{' '*18} | {results['acc']:.4f} {results['recall']:.4f} {results['f1']:.4f}"  #" - loss: {results['loss']:.4f}"
    right = results['right']
    found = results['found']
    origin = results['origin']
    info += f" {right}/{found}/{origin}"
    logger.info(info)

    info = f" Macro{' '*18} | {macro_acc:.4f} {macro_recall:.4f} {macro_f1:.4f}"  #" - loss: {results['loss']:.4f}"
    logger.info(info)

    logger.info(f"{'-' * title_len}")

    return results


def get_ner_preds_reviews(preds, test_dataflow, seg_len, seg_backoff):
    reviews = {}
    category_mentions = {}
    for json_d, tagged_text in tqdm(zip(preds, test_dataflow)):
        guid = tagged_text.guid
        text_offset = tagged_text.text_offset
        text = tagged_text.text
        entities = json_d['entities']

        #  logger.debug(f"tagged_text: {tagged_text}")
        #  logger.info(f"preds: {json_d}")

        #  logger.info(
        #      f"guid: {guid}, offset: {text_offset}, entities: {entities}")

        from ...modeling.ner_utils import LabeledText
        labeled_text = LabeledText(guid, text)
        for c, x0, x1 in entities:
            if x0 >= len(text):
                logger.info(f"x0({x0}) >= len(text)({len(text)})")
                logger.warning(
                    f"pos overflow [{guid}]:({c},{x0},{x1}) text: ({len(text)}) {text}"
                )
            if x0 < 0 or x1 < 0 or x0 > x1:
                logger.warning(f"Invalid x0({x0}) and x1({x1})")

            labeled_text.add_entity(c, x0, x1 - 1)
            if c not in category_mentions:
                category_mentions[c] = set()
            category_mentions[c].add(labeled_text.entities[-1].mention)

        annotated_text = labeled_text.get_annotated_text(seg_len, seg_backoff)

        labeled_text.offset(text_offset)
        json_data = labeled_text.to_dict()
        json_entities = json_data['entities']

        if guid not in reviews:
            reviews[guid] = {
                'guid': guid,
                'text': "",
                'annotated_text': "",
                'tags': []
            }
        reviews[guid]['text'] += text[:seg_len - seg_backoff]
        reviews[guid]['annotated_text'] += annotated_text
        reviews_tags = reviews[guid]['tags']
        for x in json_entities:
            if x not in reviews_tags:
                c = x['category']
                s = int(x['start'])
                m = x['mention']
                #  e = s + len(m) - 1
                #  m = text[s:e + 1]

                #  if not m:
                #      logger.warning(
                #          f"Mention is None. guid: {guid}, text: {text}, c: {c}, s: {s}, m: {m}"
                #      )
                reviews_tags.append({'category': c, 'start': s, 'mention': m})

        reviews_tags = sorted(reviews_tags, key=lambda x: x['start'])
        from ..utils import remove_duplicate_entities
        reviews_tags = remove_duplicate_entities(reviews_tags)
        reviews[guid]['tags'] = reviews_tags
        #  logger.info(f"guid: {guid}, tags: {reviews[guid]['tags']}")

    return reviews, category_mentions


def save_ner_preds(args, preds, test_dataflow):
    reviews, category_mentions = get_ner_preds_reviews(preds, test_dataflow,
                                                       args.seg_len,
                                                       args.seg_backoff)

    #  reviews_file = f"{args.latest_dir}/{args.dataset_name}_reviews_{args.local_id}.json"
    reviews_file = args.reviews_file

    json.dump(reviews, open(reviews_file, 'w'), ensure_ascii=False, indent=2)
    logger.info(f"Reviews file: {reviews_file}")

    category_mentions_file = f"{args.latest_dir}/{args.dataset_name}_category_mentions_{args.local_id}.txt"
    num_categories = len(category_mentions)
    num_mentions = 0
    with open(category_mentions_file, 'w') as wt:
        for c, mentions in category_mentions.items():
            for m in mentions:
                wt.write(f"{c}\t{m}\n")
                num_mentions += 1
    logger.info(
        f"Total {num_categories} categories and {num_mentions} mentions saved to {category_mentions_file}"
    )

    #  ----- Tracking -----
    #  if args.do_experiment:
    #      mlflow.log_param(f"{args.dataset_name}_reviews_file", reviews_file)
    #      mlflow.log_artifact(reviews_file, args.artifact_path)
    #      mlflow.log_param(f"{args.dataset_name}_category_mentions_file",
    #                       category_mentions_file)
    #      mlflow.log_artifact(category_mentions_file, args.artifact_path)

    return reviews_file, category_mentions_file



# -*- coding: utf-8 -*-
# File: noise.py


import numpy as np
import cv2

from .base import PhotometricAugmentor

__all__ = ['JpegNoise', 'GaussianNoise', 'SaltPepperNoise']


class JpegNoise(PhotometricAugmentor):
    """ Random JPEG noise. """

    def __init__(self, quality_range=(40, 100)):
        """
        Args:
            quality_range (tuple): range to sample JPEG quality
        """
        super(JpegNoise, self).__init__()
        self._init(locals())

    def _get_augment_params(self, img):
        return self.rng.randint(*self.quality_range)

    def _augment(self, img, q):
        enc = cv2.imencode('.jpg', img, [cv2.IMWRITE_JPEG_QUALITY, q])[1]
        return cv2.imdecode(enc, 1).astype(img.dtype)


class GaussianNoise(PhotometricAugmentor):
    """
    Add random Gaussian noise N(0, sigma^2) of the same shape to img.
    """
    def __init__(self, sigma=1, clip=True):
        """
        Args:
            sigma (float): stddev of the Gaussian distribution.
            clip (bool): clip the result to [0,255] in the end.
        """
        super(GaussianNoise, self).__init__()
        self._init(locals())

    def _get_augment_params(self, img):
        return self.rng.randn(*img.shape)

    def _augment(self, img, noise):
        old_dtype = img.dtype
        ret = img + noise * self.sigma
        if self.clip or old_dtype == np.uint8:
            ret = np.clip(ret, 0, 255)
        return ret.astype(old_dtype)


class SaltPepperNoise(PhotometricAugmentor):
    """ Salt and pepper noise.
        Randomly set some elements in image to 0 or 255, regardless of its channels.
    """

    def __init__(self, white_prob=0.05, black_prob=0.05):
        """
        Args:
            white_prob (float), black_prob (float): probabilities setting an element to 255 or 0.
        """
        assert white_prob + black_prob <= 1, "Sum of probabilities cannot be greater than 1"
        super(SaltPepperNoise, self).__init__()
        self._init(locals())

    def _get_augment_params(self, img):
        return self.rng.uniform(low=0, high=1, size=img.shape)

    def _augment(self, img, param):
        img[param > (1 - self.white_prob)] = 255
        img[param < self.black_prob] = 0
        return img



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os, re, json

import numpy as np
from loguru import logger
from tqdm import tqdm
import rich
from collections import defaultdict


def get_module(module_name):
    import importlib
    module_cls = importlib.import_module(module_name)

    return module_cls


def main():
    import sys
    module_name = sys.argv[1]
    module_cls = get_module(module_name)

    train_data_generator = None
    if 'train_data_generator' in module_cls.__dict__:
        train_data_generator = module_cls.train_data_generator
    test_data_generator = None
    if 'test_data_generator' in module_cls.__dict__:
        test_data_generator = module_cls.test_data_generator

    is_ner = False
    is_glue = False
    if 'ner_labels' in module_cls.__dict__:
        is_ner = True
    if 'glue_labels' in module_cls.__dict__:
        is_glue = True

    if train_data_generator is not None:
        if is_glue:
            train_samples = [x for x in train_data_generator()]
            labels = [label for _, _, _, label in train_samples]
            num_train_samples = len(labels)

            text1_lens = [len(text_a) for _, text_a, _, _ in train_samples]
            len_mean1 = np.mean(text1_lens)
            len_std1 = np.std(text1_lens)
            len_min1 = np.min(text1_lens)
            len_max1 = np.max(text1_lens)

            text2_lens = [
                len(text_b) if text_b is not None else 0
                for _, _, text_b, _ in train_samples
            ]
            len_mean2 = np.mean(text2_lens)
            len_std2 = np.std(text2_lens)
            len_min2 = np.min(text2_lens)
            len_max2 = np.max(text2_lens)

            labels = sorted(list(set(labels)))
            rich.print(f"{len(labels)} glue_labels: {labels}")
            rich.print(f"num_train_samples: {num_train_samples}")
            rich.print(f"text_a lengths:")
            rich.print(
                f"mean: {len_mean1:.2f}, std: {len_std1:.2f}, min: {len_min1}, max: {len_max1}"
            )
            rich.print(f"text_b lengths:")
            rich.print(
                f"mean: {len_mean2:.2f}, std: {len_std2:.2f}, min: {len_min2}, max: {len_max2}"
            )
            # 标签分布情况
            labels_map = defaultdict(int)
            for sample in train_samples:
                c = sample[3]
                labels_map[c] += 1
            rich.print(f"{labels_map}")
        elif is_ner:
            train_samples = [x for x in train_data_generator()]
            all_tags = [tags for _, _, _, tags in train_samples]
            num_train_samples = len(all_tags)

            labels = []
            for tags in all_tags:
                for tag in tags:
                    c = tag['category']
                    if c not in labels:
                        labels.append(c)

            text_lens = [len(text) for _, text, _, _ in train_samples]
            len_mean = np.mean(text_lens)
            len_std = np.std(text_lens)
            len_min = np.min(text_lens)
            len_max = np.max(text_lens)

            labels = sorted(list(set(labels)))
            rich.print(f"{len(labels)} ner labels: {labels}")
            rich.print(f"num_train_samples: {num_train_samples}")
            rich.print(f"text lengths:")
            rich.print(
                f"mean: {len_mean:.2f}, std: {len_std:.2f}, min: {len_min}, max: {len_max}"
            )

    if test_data_generator is not None:
        if is_glue:
            test_samples = [x for x in test_data_generator()]
            text1_lens = [len(text_a) for _, text_a, _, _ in test_samples]
            len_mean1 = np.mean(text1_lens)
            len_std1 = np.std(text1_lens)
            len_min1 = np.min(text1_lens)
            len_max1 = np.max(text1_lens)

            text2_lens = [
                len(text_b) if text_b is not None else 0
                for _, _, text_b, _ in test_samples
            ]
            len_mean2 = np.mean(text2_lens)
            len_std2 = np.std(text2_lens)
            len_min2 = np.min(text2_lens)
            len_max2 = np.max(text2_lens)
            num_test_samples = len(test_samples)
            rich.print(f"num_test_samples: {num_test_samples}")
            rich.print(f"text_a lengths:")
            rich.print(
                f"mean: {len_mean1:.2f}, std: {len_std1:.2f}, min: {len_min1}, max: {len_max1}"
            )
            rich.print(f"text_b lengths:")
            rich.print(
                f"mean: {len_mean2:.2f}, std: {len_std2:.2f}, min: {len_min2}, max: {len_max2}"
            )
        elif is_ner:
            test_samples = [x for x in test_data_generator()]
            text_lens = [len(text) for _, text, _, _ in test_samples]
            num_test_samples = len(text_lens)

            len_mean = np.mean(text_lens)
            len_std = np.std(text_lens)
            len_min = np.min(text_lens)
            len_max = np.max(text_lens)

            rich.print(f"num_test_samples: {num_test_samples}")
            rich.print(f"text lengths:")
            rich.print(
                f"mean: {len_mean:.2f}, std: {len_std:.2f}, min: {len_min}, max: {len_max}"
            )


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from loguru import logger


def get_ner_results(metric):
    eval_info, entity_info = metric.result()
    results = {f'{key}': value for key, value in eval_info.items()}

    title = f"{' '*24}   acc    recall f1    "
    title += " Right/Pred/True "
    title_len = len(title.encode('gbk'))
    logger.info(f"{'=' * title_len}")
    logger.info(title)
    logger.info(f"{'-' * title_len}")

    sorted_entity_info = sorted(entity_info.items(),
                                key=lambda x: x[1]['f1'],
                                reverse=True)
    total_acc = 0
    total_recall = 0
    total_f1 = 0
    for key, metrics in sorted_entity_info:
        if ':' in key:
            category = key.split(':')[1]
        else:
            category = key
        if metric.ignore_categories and category in metric.ignore_categories:
            disp_key = ' -' + key[:16]
        else:
            disp_key = ' ' + key[:16]
        disp_key += ' ' * (24 - len(disp_key.encode('gbk')))
        info = f"{disp_key} | {metrics['acc']:.4f} {metrics['recall']:.4f} {metrics['f1']:.4f}"
        right = metrics['right']
        found = metrics['found']
        origin = metrics['origin']
        info += f" {right}/{found}/{origin}"
        logger.info(info)
        if metric.ignore_categories and category in metric.ignore_categories:
            pass
        else:
            total_acc += metrics['acc']
            total_recall += metrics['recall']
            total_f1 += metrics['f1']
    if metric.ignore_categories:
        num_categories = len(sorted_entity_info) - len(
            metric.ignore_categories)
    else:
        num_categories = len(sorted_entity_info)
    macro_acc = total_acc / num_categories
    macro_recall = total_recall / num_categories
    macro_f1 = total_f1 / num_categories
    logger.info(f"{'-' * title_len}")

    info = f" Micro{' '*18} | {results['acc']:.4f} {results['recall']:.4f} {results['f1']:.4f}"  #" - loss: {results['loss']:.4f}"
    right = results['right']
    found = results['found']
    origin = results['origin']
    info += f" {right}/{found}/{origin}"
    logger.info(info)

    info = f" Macro{' '*18} | {macro_acc:.4f} {macro_recall:.4f} {macro_f1:.4f}"  #" - loss: {results['loss']:.4f}"
    logger.info(info)

    logger.info(f"{'-' * title_len}")

    return results



# -*- coding: utf-8 -*-
# File: caltech101.py


import os

from ...utils import logger
from ...utils.fs import download, get_dataset_path
from ..base import RNGDataFlow

__all__ = ["Caltech101Silhouettes"]


def maybe_download(url, work_directory):
    """Download the data from Marlin's website, unless it's already here."""
    filename = url.split("/")[-1]
    filepath = os.path.join(work_directory, filename)
    if not os.path.exists(filepath):
        logger.info("Downloading to {}...".format(filepath))
        download(url, work_directory)
    return filepath


class Caltech101Silhouettes(RNGDataFlow):
    """
    Produces [image, label] in Caltech101 Silhouettes dataset,
    image is 28x28 in the range [0,1], label is an int in the range [0,100].
    """

    _DIR_NAME = "caltech101_data"
    _SOURCE_URL = "https://people.cs.umass.edu/~marlin/data/"

    def __init__(self, name, shuffle=True, dir=None):
        """
        Args:
            name (str): 'train', 'test', 'val'
            shuffle (bool): shuffle the dataset
        """
        if dir is None:
            dir = get_dataset_path(self._DIR_NAME)
        assert name in ['train', 'test', 'val']
        self.name = name
        self.shuffle = shuffle

        def get_images_and_labels(data_file):
            f = maybe_download(self._SOURCE_URL + data_file, dir)
            data = scipy.io.loadmat(f)
            return data

        self.data = get_images_and_labels("caltech101_silhouettes_28_split1.mat")

        if self.name == "train":
            self.images = self.data["train_data"].reshape((4100, 28, 28))
            self.labels = self.data["train_labels"].ravel() - 1
        elif self.name == "test":
            self.images = self.data["test_data"].reshape((2307, 28, 28))
            self.labels = self.data["test_labels"].ravel() - 1
        else:
            self.images = self.data["val_data"].reshape((2264, 28, 28))
            self.labels = self.data["val_labels"].ravel() - 1

    def __len__(self):
        return self.images.shape[0]

    def __iter__(self):
        idxs = list(range(self.__len__()))
        if self.shuffle:
            self.rng.shuffle(idxs)
        for k in idxs:
            img = self.images[k]
            label = self.labels[k]
            yield [img, label]


try:
    import scipy.io
except ImportError:
    from ...utils.develop import create_dummy_class
    Caltech101Silhouettes = create_dummy_class('Caltech101Silhouettes', 'scipy.io') # noqa


if __name__ == "__main__":
    ds = Caltech101Silhouettes("train")
    ds.reset_state()
    for _ in ds:
        from IPython import embed

        embed()
        break



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import copy
import json
from collections import defaultdict
from pathlib import Path

import numpy as np
import torch
from loguru import logger
from torch.utils.data import (DataLoader, RandomSampler, SequentialSampler,
                              TensorDataset)
from tqdm import tqdm

from ...utils import match_tokenized_to_untokenized, seg_generator
from ...utils.multiprocesses import (barrier_leader_process,
                                     barrier_member_processes,
                                     is_multi_processes)
from ..spo_utils import InputExample


class InputFeature(object):
    """A single set of features of data."""
    def __init__(self, input_ids, input_mask, input_len, segment_ids,
                 sub_labels, seed_sub, obj_labels, token_offsets, text,
                 spo_list):
        self.input_ids = input_ids
        self.input_mask = input_mask
        self.segment_ids = segment_ids
        self.input_len = input_len
        self.sub_labels = sub_labels
        self.seed_sub = seed_sub
        self.obj_labels = obj_labels
        self.token_offsets = token_offsets
        self.text = text
        self.spo_list = spo_list

    def __repr__(self):
        return str(self.to_json_string())

    def __eq__(self, other):
        return self.input_ids == other.input_ids and \
            self.input_mask == other.input_mask and \
            self.segment_ids == other.segment_ids and \
            self.labels == other.labels and \
            self.sub_labels == other.sub_labels and \
            self.seed_sub == other.seed_sub and \
            self.obj_labels == other.obj_labels and \
            self.input_len == other.input_len and \
            self.token_offsets == other.token_offsets

    def to_dict(self):
        """Serializes this instance to a Python dictionary."""
        output = copy.deepcopy(self.__dict__)
        return output

    def to_json_string(self):
        """Serializes this instance to a JSON string."""
        return json.dumps(self.to_dict(), indent=2, sort_keys=True) + "\n"


def batch_to_input_data(batch):
    all_input_ids = torch.stack([e.input_ids for e in batch])
    all_input_mask = torch.stack([e.input_mask for e in batch])
    all_segment_ids = torch.stack([e.segment_ids for e in batch])
    all_sub_labels = torch.stack([e.sub_labels for e in batch])
    all_seed_subs = torch.stack([e.seed_sub for e in batch])
    all_obj_labels = torch.stack([e.obj_labels for e in batch])
    all_lens = torch.stack([e.input_len for e in batch])
    all_token_offsets = torch.stack([e.token_offsets for e in batch])
    all_texts = [e.text for e in batch]
    all_spo_lists = [e.spo_list for e in batch]

    return (all_input_ids, all_input_mask, all_segment_ids, all_sub_labels,
            all_seed_subs, all_obj_labels, all_lens, all_token_offsets,
            all_texts, all_spo_lists)


def encode_examples(examples, label2id, tokenizer, max_seq_length):

    num_labels = len(label2id)

    #  texts = [e.text for e in examples]
    #  all_tokens = [['[CLS]'] + tokenizer.encode(
    #      text, add_special_tokens=False).tokens[:max_seq_length - 2] +
    #                ['[SEP]'] for text in tqdm(texts, desc="Tokenize")]
    #  all_token2texts = [
    #      match_tokenized_to_untokenized(tokens, text)
    #      for tokens, text in zip(all_tokens, texts)
    #  ]
    #  all_input_ids = [
    #      tokenizer.convert_tokens_to_ids(tokens) for tokens in all_tokens
    #  ]
    #  all_attention_mask = [[1] * len(input_ids) for input_ids in all_input_ids]
    #  all_token_type_ids = [[0] * len(input_ids) for input_ids in all_input_ids]

    for e in examples:
        assert len(e.text) <= max_seq_length - 2
        #  logger.warning(f"text: {e.text}")
        #  logger.warning(f"spo_list: {e.spo_list}")
    texts = [e.text[:max_seq_length - 2] for e in examples]
    all_encodes = [
        tokenizer.encode(text, add_special_tokens=True)
        for text in tqdm(texts, desc="Tokenize")
    ]

    #  all_tokens = [encode.tokens for encode in all_encodes]
    #  all_token_offsets = [encode.offsets for encode in all_encodes]
    #  all_input_ids = [encode.ids for encode in all_encodes]
    #  all_attention_mask = [encode.attention_mask for encode in all_encodes]
    #  all_token_type_ids = [encode.type_ids for encode in all_encodes]

    all_encodes = tokenizer.batch_encode(texts, add_special_tokens=True)
    all_tokens = all_encodes['tokens']
    all_token_offsets = all_encodes['offsets']
    all_token2char = all_encodes['token2char']
    all_char2token = all_encodes['char2token']

    all_input_ids = all_encodes['ids']
    all_attention_mask = all_encodes['attention_mask']
    all_token_type_ids = all_encodes['type_ids']

    all_input_lens = [len(tokens) for tokens in all_tokens]

    all_padding_lens = [max_seq_length - n for n in all_input_lens]
    for i, (input_ids, attention_mask, token_type_ids, token2char,
            token_offsets, padding_length) in enumerate(
                zip(all_input_ids, all_attention_mask, all_token_type_ids,
                    all_token2char, all_token_offsets, all_padding_lens)):
        all_input_ids[i] = input_ids + [0] * padding_length
        all_attention_mask[i] = attention_mask + [0] * padding_length
        all_token_type_ids[i] = token_type_ids + [0] * padding_length
        all_token2char[i] = token2char + [0] * padding_length
        all_token_offsets[i] = token_offsets + [(0, 0)] * padding_length

    def encode_spo(spo_list, num_tokens, text_len, char2token):
        """
        spo_list: [(s, p, o), ...]
            s: (sub_s, sub_e)
            p: predicate
            o: (obj_s, obj_e)
        #  spo_list: [spo1, spo2, ...]
        #      spo: {  'subject':   {'category': sub_c, 'start': sub_s, 'mention': sub_m},
        #              'predicate': predicate,
        #              'object':    {'category': obj_c, 'start': obj_s, 'mention': obj_m}
        #          }

        """
        # 将spo三元组列表转换为词典{(s0, s1): [(o_0, o_1, p)]}
        spoes = defaultdict(list)
        if spo_list:
            for s, p, o in spo_list:
                s_start, s_mention = s
                s_end = s_start + len(s_mention) - 1
                #  assert s_end <= text_len, f"Skip spo. s_end >= {text_len}: {s}, {p}, {o}. "

                o_start, o_mention = o
                o_end = o_start + len(o_mention) - 1
                #  assert o_end <= text_len, f"Skip spo. o_end >= {text_len}: {s}, {p}, {o}. "

                #  logger.warning(f"s: {s_start}:{s_end}, o: {o_start}:{o_end}")
                #  logger.info(f"char2toke: {char2token}")
                s_start = char2token[s_start]
                assert s_start >= 0 and s_start < num_tokens - 2, f"s_start overflow. |{s}|{p}|{o}| ({s_start}:{s_end}, num_tokens: {num_tokens})"
                s_end = char2token[s_end]
                assert s_end >= 0 and s_end < num_tokens - 2, f"s_end overflow. |{s}|{p}|{o}| ({s_start}:{s_end}, num_tokens: {num_tokens})"
                o_start = char2token[o_start]
                assert o_start >= 0 and o_start < num_tokens - 2, f"o_start overflow. |{s}|{p}|{o}| ({o_start}:{o_end}, num_tokens: {num_tokens})"
                o_end = char2token[o_end]
                assert o_end >= 0 and o_end < num_tokens - 2, f"o_end overflow. |{s}|{p}|{o}| ({o_start}:{o_end}, num_tokens: {num_tokens})"

                spoes[(s_start, s_end)].append((o_start, o_end, label2id[p]))

        sub_labels = np.zeros((num_tokens, 2), dtype=int)
        # 对应的object标签
        obj_labels = np.zeros((num_tokens, num_labels, 2), dtype=int)
        #  logger.info(f"obj_labels.shape: {obj_labels.shape}")
        seed_sub = (0, 0)
        if spoes:

            for s in spoes:
                sub_labels[s[0] + 1, 0] = 1
                sub_labels[s[1] + 1, 1] = 1

            #  logger.info(f"spoes: {spoes}")
            # 随机选一个subject
            start, end = np.array(list(spoes.keys())).T
            start = np.random.choice(start)
            end = np.random.choice(end[end >= start])
            seed_sub = (start, end)

            for o in spoes.get(seed_sub, []):
                #  logger.info(f"o: {o}")
                obj_labels[o[0] + 1, o[2], 0] = 1
                obj_labels[o[1] + 1, o[2], 1] = 1

        return sub_labels, seed_sub, obj_labels

    all_sub_labels = []
    all_seed_subs = []
    all_obj_labels = []

    all_spo_lists = [e.spo_list for e in examples]
    for example, input_ids, attention_mask, token_type_ids, spo_list, input_len, char2token in tqdm(
            zip(examples, all_input_ids, all_attention_mask,
                all_token_type_ids, all_spo_lists, all_input_lens,
                all_char2token)):
        num_tokens = len(input_ids)
        #  logger.info(f"example.text: {example.text}")
        sub_labels, seed_sub, obj_labels = encode_spo(spo_list, num_tokens,
                                                      input_len - 2,
                                                      char2token)
        assert len(input_ids) == max_seq_length
        assert len(attention_mask) == max_seq_length
        assert len(token_type_ids) == max_seq_length
        assert len(sub_labels) == max_seq_length
        assert len(obj_labels) == max_seq_length

        all_sub_labels.append(sub_labels)
        all_seed_subs.append(seed_sub)
        all_obj_labels.append(obj_labels)
        #  logger.info(
        #      f"obj_labels: {torch.tensor(obj_labels, dtype=torch.long).cuda()}")

    logger.debug(f"all_input_ids.shape: {np.array(all_input_ids).shape}")
    logger.debug(
        f"all_attention_mask.shape: {np.array(all_attention_mask).shape}")
    logger.debug(
        f"all_token_type_ids.shape: {np.array(all_token_type_ids).shape}")
    logger.debug(f"all_sub_labels.shape: {np.array(all_sub_labels).shape}")
    logger.debug(f"all_obj_labels.shape: {np.array(all_obj_labels).shape}")
    logger.debug(f"all_input_lens.shape: {np.array(all_input_lens).shape}")
    logger.debug(
        f"all_token_offsets.shape: {np.array(all_token_offsets).shape}")
    assert np.array(all_input_ids).shape[
        1] == max_seq_length, f"shape: {np.array(all_input_ids).shape}"
    assert np.array(all_attention_mask).shape[1] == max_seq_length
    assert np.array(all_token_type_ids).shape[1] == max_seq_length

    for (start, end), text_len in zip(all_seed_subs, all_input_lens):
        assert start < text_len and end < text_len, f"start: {start}, end: {end}, text_len: {text_len}"

    all_input_ids = torch.from_numpy(np.array(all_input_ids, dtype=np.int64))
    all_attention_mask = torch.from_numpy(
        np.array(all_attention_mask, dtype=np.int64))
    all_token_type_ids = torch.from_numpy(
        np.array(all_token_type_ids, dtype=np.int64))
    all_sub_labels = torch.from_numpy(np.array(all_sub_labels, dtype=np.int64))
    all_seed_subs = torch.from_numpy(np.array(all_seed_subs, dtype=np.int64))
    all_obj_labels = torch.from_numpy(np.array(all_obj_labels, dtype=np.int64))
    all_input_lens = torch.from_numpy(np.array(all_input_lens, dtype=np.int64))
    all_token_offsets = torch.from_numpy(
        np.array(all_token_offsets, dtype=np.int64))

    all_features = [
        InputFeature(input_ids=input_ids,
                     input_mask=attention_mask,
                     segment_ids=token_type_ids,
                     sub_labels=sub_labels,
                     seed_sub=seed_subs,
                     obj_labels=obj_labels,
                     input_len=input_len,
                     token_offsets=token_offsets,
                     text=text,
                     spo_list=spo_list)
        for input_ids, attention_mask, token_type_ids, sub_labels, seed_subs,
        obj_labels, input_len, token_offsets, text, spo_list in tqdm(
            zip(all_input_ids, all_attention_mask, all_token_type_ids,
                all_sub_labels, all_seed_subs, all_obj_labels, all_input_lens,
                all_token_offsets, texts, all_spo_lists))
    ]
    logger.warning(f"Prepare data done.")
    #  logger.info(f"all_token_offsets[:1]: {all_token_offsets[:1]}")
    #  logger.info(f"{np.array(all_token_offsets).shape}")
    #  logger.info(f"{np.array(all_token_offsets[:1]).shape}")
    #  logger.info(f"{np.array(all_token_offsets[:2]).shape}")

    return all_features

    #  return (all_features, {
    #      'input_ids': all_input_ids,
    #      'attention_mask': all_attention_mask,
    #      'token_type_ids': all_token_type_ids,
    #      'sub_labels': all_sub_labels,
    #      "seed_sub": all_seed_subs,
    #      'obj_labels': all_obj_labels,
    #      "input_lens": all_input_lens,
    #      "token_offsets": all_token_offsets,
    #  })

    #  return (all_features, {
    #  'input_ids':
    #  torch.from_numpy(np.array(all_input_ids, dtype=np.int64)),
    #  'attention_mask':
    #  torch.from_numpy(np.array(all_attention_mask, dtype=np.int64)),
    #  'token_type_ids':
    #  torch.from_numpy(np.array(all_token_type_ids, dtype=np.int64)),
    #  'sub_labels':
    #  torch.from_numpy(np.array(all_sub_labels, dtype=np.int64)),
    #  "seed_sub":
    #  torch.from_numpy(np.array(all_seed_subs, dtype=np.int64)),
    #  'obj_labels':
    #  torch.from_numpy(np.array(all_obj_labels, dtype=np.int64)),
    #  "input_lens":
    #  torch.from_numpy(np.array(all_input_lens, dtype=np.int64)),
    #  "token_offsets":
    #  torch.from_numpy(np.array(all_token_offsets, dtype=np.int64)),
    #  })


#

#  def examples_to_dataset(examples, label2id, tokenizer, max_seq_length):
#      all_features, outputs = encode_examples(examples, label2id, tokenizer,
#                                              max_seq_length)
#
#      logger.warning(f"Buid TensorDataset")
#      dataset = TensorDataset(
#          outputs['input_ids'],
#          outputs['attention_mask'],
#          outputs['token_type_ids'],
#          outputs['sub_labels'],
#          outputs['seed_sub'],
#          outputs['obj_labels'],
#          outputs['input_lens'],
#          outputs['token_offsets'],
#      )
#
#      assert len(dataset) == len(all_features)
#
#      return dataset, all_features



# -*- coding: utf-8 -*-
# File: crop.py

import numpy as np
import cv2

from ...utils.argtools import shape2d
from ...utils.develop import log_deprecated
from .base import ImageAugmentor, ImagePlaceholder
from .transform import CropTransform, TransformList, ResizeTransform, PhotometricTransform
from .misc import ResizeShortestEdge

__all__ = ['RandomCrop', 'CenterCrop', 'RandomCropRandomShape',
           'GoogleNetRandomCropAndResize', 'RandomCutout']


class RandomCrop(ImageAugmentor):
    """ Randomly crop the image into a smaller one """

    def __init__(self, crop_shape):
        """
        Args:
            crop_shape: (h, w), int or a tuple of int
        """
        crop_shape = shape2d(crop_shape)
        crop_shape = (int(crop_shape[0]), int(crop_shape[1]))
        super(RandomCrop, self).__init__()
        self._init(locals())

    def get_transform(self, img):
        orig_shape = img.shape
        assert orig_shape[0] >= self.crop_shape[0] \
            and orig_shape[1] >= self.crop_shape[1], orig_shape
        diffh = orig_shape[0] - self.crop_shape[0]
        h0 = self.rng.randint(diffh + 1)
        diffw = orig_shape[1] - self.crop_shape[1]
        w0 = self.rng.randint(diffw + 1)
        return CropTransform(h0, w0, self.crop_shape[0], self.crop_shape[1])


class CenterCrop(ImageAugmentor):
    """ Crop the image at the center"""

    def __init__(self, crop_shape):
        """
        Args:
            crop_shape: (h, w) tuple or a int
        """
        crop_shape = shape2d(crop_shape)
        self._init(locals())

    def get_transform(self, img):
        orig_shape = img.shape
        assert orig_shape[0] >= self.crop_shape[0] \
            and orig_shape[1] >= self.crop_shape[1], orig_shape
        h0 = int((orig_shape[0] - self.crop_shape[0]) * 0.5)
        w0 = int((orig_shape[1] - self.crop_shape[1]) * 0.5)
        return CropTransform(h0, w0, self.crop_shape[0], self.crop_shape[1])


class RandomCropRandomShape(ImageAugmentor):
    """ Random crop with a random shape"""

    def __init__(self, wmin, hmin,
                 wmax=None, hmax=None,
                 max_aspect_ratio=None):
        """
        Randomly crop a box of shape (h, w), sampled from [min, max] (both inclusive).
        If max is None, will use the input image shape.

        Args:
            wmin, hmin, wmax, hmax: range to sample shape.
            max_aspect_ratio (float): this argument has no effect and is deprecated.
        """
        super(RandomCropRandomShape, self).__init__()
        if max_aspect_ratio is not None:
            log_deprecated("RandomCropRandomShape(max_aspect_ratio)", "It is never implemented!", "2020-06-06")
        self._init(locals())

    def get_transform(self, img):
        hmax = self.hmax or img.shape[0]
        wmax = self.wmax or img.shape[1]
        h = self.rng.randint(self.hmin, hmax + 1)
        w = self.rng.randint(self.wmin, wmax + 1)
        diffh = img.shape[0] - h
        diffw = img.shape[1] - w
        assert diffh >= 0 and diffw >= 0, str(diffh) + ", " + str(diffw)
        y0 = 0 if diffh == 0 else self.rng.randint(diffh)
        x0 = 0 if diffw == 0 else self.rng.randint(diffw)
        return CropTransform(y0, x0, h, w)


class GoogleNetRandomCropAndResize(ImageAugmentor):
    """
    The random crop and resize augmentation proposed in
    Sec. 6 of "Going Deeper with Convolutions" by Google.
    This implementation follows the details in ``fb.resnet.torch``.

    It attempts to crop a random rectangle with 8%~100% area of the original image,
    and keep the aspect ratio between 3/4 to 4/3. Then it resize this crop to the target shape.
    If such crop cannot be found in 10 iterations, it will do a ResizeShortestEdge + CenterCrop.
    """
    def __init__(self, crop_area_fraction=(0.08, 1.),
                 aspect_ratio_range=(0.75, 1.333),
                 target_shape=224, interp=cv2.INTER_LINEAR):
        """
        Args:
            crop_area_fraction (tuple(float)): Defaults to crop 8%-100% area.
            aspect_ratio_range (tuple(float)): Defaults to make aspect ratio in 3/4-4/3.
            target_shape (int): Defaults to 224, the standard ImageNet image shape.
        """
        super(GoogleNetRandomCropAndResize, self).__init__()
        self._init(locals())

    def get_transform(self, img):
        h, w = img.shape[:2]
        area = h * w
        for _ in range(10):
            targetArea = self.rng.uniform(*self.crop_area_fraction) * area
            aspectR = self.rng.uniform(*self.aspect_ratio_range)
            ww = int(np.sqrt(targetArea * aspectR) + 0.5)
            hh = int(np.sqrt(targetArea / aspectR) + 0.5)
            if self.rng.uniform() < 0.5:
                ww, hh = hh, ww
            if hh <= h and ww <= w:
                x1 = self.rng.randint(0, w - ww + 1)
                y1 = self.rng.randint(0, h - hh + 1)
                return TransformList([
                    CropTransform(y1, x1, hh, ww),
                    ResizeTransform(hh, ww, self.target_shape, self.target_shape, interp=self.interp)
                ])
        resize = ResizeShortestEdge(self.target_shape, interp=self.interp).get_transform(img)
        out_shape = (resize.new_h, resize.new_w)
        crop = CenterCrop(self.target_shape).get_transform(ImagePlaceholder(shape=out_shape))
        return TransformList([resize, crop])


class RandomCutout(ImageAugmentor):
    """
    The cutout augmentation, as described in https://arxiv.org/abs/1708.04552
    """
    def __init__(self, h_range, w_range, fill=0.):
        """
        Args:
            h_range (int or tuple): the height of rectangle to cut.
                If a tuple, will randomly sample from this range [low, high)
            w_range (int or tuple): similar to above
            fill (float): the fill value
        """
        super(RandomCutout, self).__init__()
        self._init(locals())

    def _get_cutout_shape(self):
        if isinstance(self.h_range, int):
            h = self.h_range
        else:
            h = self.rng.randint(self.h_range)

        if isinstance(self.w_range, int):
            w = self.w_range
        else:
            w = self.rng.randint(self.w_range)
        return h, w

    @staticmethod
    def _cutout(img, y0, x0, h, w, fill):
        img[y0:y0 + h, x0:x0 + w] = fill
        return img

    def get_transform(self, img):
        h, w = self._get_cutout_shape()
        x0 = self.rng.randint(0, img.shape[1] + 1 - w)
        y0 = self.rng.randint(0, img.shape[0] + 1 - h)
        return PhotometricTransform(
            lambda img: RandomCutout._cutout(img, y0, x0, h, w, self.fill),
            "cutout")



# -*- coding: utf-8 -*-
# File: raw.py


import copy
import numpy as np
import six

from .base import DataFlow, RNGDataFlow

__all__ = ['FakeData', 'DataFromQueue', 'DataFromList', 'DataFromGenerator', 'DataFromIterable']


class FakeData(RNGDataFlow):
    """ Generate fake data of given shapes"""

    def __init__(self, shapes, size=1000, random=True, dtype='float32', domain=(0, 1)):
        """
        Args:
            shapes (list): a list of lists/tuples. Shapes of each component.
            size (int): size of this DataFlow.
            random (bool): whether to randomly generate data every iteration.
                Note that merely generating the data could sometimes be time-consuming!
            dtype (str or list): data type as string, or a list of data types.
            domain (tuple or list): (min, max) tuple, or a list of such tuples
        """
        super(FakeData, self).__init__()
        self.shapes = shapes
        self._size = int(size)
        self.random = random
        self.dtype = [dtype] * len(shapes) if isinstance(dtype, six.string_types) else dtype
        self.domain = [domain] * len(shapes) if isinstance(domain, tuple) else domain
        assert len(self.dtype) == len(self.shapes)
        assert len(self.domain) == len(self.domain)

    def __len__(self):
        return self._size

    def __iter__(self):
        if self.random:
            for _ in range(self._size):
                val = []
                for k in range(len(self.shapes)):
                    v = self.rng.rand(*self.shapes[k]) * (self.domain[k][1] - self.domain[k][0]) + self.domain[k][0]
                    val.append(v.astype(self.dtype[k]))
                yield val
        else:
            val = []
            for k in range(len(self.shapes)):
                v = self.rng.rand(*self.shapes[k]) * (self.domain[k][1] - self.domain[k][0]) + self.domain[k][0]
                val.append(v.astype(self.dtype[k]))
            for _ in range(self._size):
                yield copy.copy(val)


class DataFromQueue(DataFlow):
    """ Produce data from a queue """
    def __init__(self, queue):
        """
        Args:
            queue (queue): a queue with ``get()`` method.
        """
        self.queue = queue

    def __iter__(self):
        while True:
            yield self.queue.get()


class DataFromList(RNGDataFlow):
    """ Wrap a list of datapoints to a DataFlow"""

    def __init__(self, lst, shuffle=True):
        """
        Args:
            lst (list): input list. Each element is a datapoint.
            shuffle (bool): shuffle data.
        """
        super(DataFromList, self).__init__()
        self.lst = lst
        self.shuffle = shuffle

    def __len__(self):
        return len(self.lst)

    def __iter__(self):
        if not self.shuffle:
            yield from self.lst
        else:
            idxs = np.arange(len(self.lst))
            self.rng.shuffle(idxs)
            for k in idxs:
                yield self.lst[k]


class DataFromGenerator(DataFlow):
    """
    Wrap a generator to a DataFlow.
    The dataflow will not have length.
    """
    def __init__(self, gen):
        """
        Args:
            gen: iterable, or a callable that returns an iterable
        """
        self._gen = gen

    def __iter__(self):
        if not callable(self._gen):
            yield from self._gen
        else:
            yield from self._gen()

    def __len__(self):
        return len(self._gen)


class DataFromIterable(DataFlow):
    """ Wrap an iterable of datapoints to a DataFlow"""
    def __init__(self, iterable):
        """
        Args:
            iterable: an iterable object
        """
        self._itr = iterable
        try:
            self._len = len(iterable)
        except Exception:
            self._len = None

    def __len__(self):
        if self._len is None:
            raise NotImplementedError
        return self._len

    def __iter__(self):
        yield from self._itr



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os, shutil, random, json
from dataclasses import dataclass, field
from typing import List
from tqdm import tqdm
from loguru import logger
#  import mlflow


class Params:
    def log(self):
        pass
        #  for k, v in self.__dict__.items():
        #      if isinstance(v, Params):
        #          v.log()
        #      else:
        #          mlflow.log_param(k, v)

    def debug(self):
        for k, v in sorted(self.__dict__.items()):
            if isinstance(v, Params):
                v.debug()
            else:
                logger.debug(f"{k}: {v}")

    def update_parser(self, parser):
        for k, v in self.__dict__.items():
            if isinstance(v, Params):
                v.update_parser(parser)
            else:
                parser.set_defaults(**{k: v})
        return parser

    def update_args(self, args):
        for k, v in self.__dict__.items():
            if isinstance(v, Params):
                v.update_args(args)
            else:
                setattr(args, k, v)

        return args


@dataclass
class CommonParams(Params):
    dataset_name: str = None
    experiment_name: str = None
    tracking_uri: str = None
    artifact_path: str = None
    train_file: str = None
    eval_file: str = None
    test_file: str = None
    learning_rate: float = 2e-5
    train_max_seq_length: int = 256
    eval_max_seq_length: int = 256
    predict_max_seq_length: int = 256
    per_gpu_train_batch_size: int = 16
    per_gpu_eval_batch_size: int = 16
    per_gpu_predict_batch_size: int = 16
    seg_len: int = 254
    seg_backoff: int = 64
    num_train_epochs: int = 5
    fold: int = 0
    num_augments: int = 0
    enable_kd: bool = False
    kd_coeff: float = 1.0
    kd_decay: float = 0.995
    enable_sda: bool = False
    sda_teachers: int = 3
    sda_stategy: str = "recent_models"
    sda_empty_first: bool = False
    sda_coeff: float = 1.0
    sda_decay: float = 0.995
    loss_type: str = "CrossEntropyLoss"
    focalloss_gamma: float = 1.5
    focalloss_alpha: List = None
    diceloss_weight: List = None
    model_type: str = "bert"
    model_path: str = None
    train_rate: float = 0.9
    train_sample_rate: float = 1.0
    fp16: bool = True
    seed: int = 8864
    best_index: str = "f1"
    random_type: str = None
    is_english: bool = False
    allow_overlap: bool = False
    adam_epsilon: float = 1e-8
    weight_decay: float = 0.0
    max_train_examples: int = 0
    confidence: float = 0.35
    enable_nested_entities: bool = False
    emotion_words_file: str = None
    cc: str = None
    brat_data_dir: str = None
    aug_train_only: bool = False

    def __post_init__(self):
        tracking_uri = self.tracking_uri
        if 'TRACKING_URI' in os.environ:
            tracking_uri = os.environ['TRACKING_URI']
        self.tracking_uri = tracking_uri

        artifact_path = self.artifact_path
        if 'ARTIFACT_PATH' in os.environ:
            artifact_path = os.environ['ARTIFACT_PATH']
        self.artifact_path = artifact_path


@dataclass
class NerParams(Params):
    ner_labels: List[str] = field(default_factory=list)
    ner_type: str = "span"
    no_crf_loss: bool = False
    soft_label: bool = False
    ignore_categories: List[str] = None


@dataclass
class SpoParams(Params):
    predicate_labels: List[str] = field(default_factory=list)


@dataclass
class GlueParams(Params):
    glue_labels: List[str] = field(default_factory=list)


@dataclass
class NerAppParams(Params):
    common_params: CommonParams = field(default_factory=CommonParams)
    ner_params: NerParams = field(default_factory=NerParams)


@dataclass
class SpoAppParams(Params):
    common_params: CommonParams = field(default_factory=CommonParams)
    spo_params: SpoParams = field(default_factory=SpoParams)


@dataclass
class GlueAppParams(Params):
    common_params: CommonParams = field(default_factory=CommonParams)
    glue_params: GlueParams = field(default_factory=GlueParams)


def log_global_params(args, experiment_params):
    pass
    #  if args.do_experiment:
    #      run_id = mlflow.active_run().info.run_id
    #      mlflow.log_param("run_id", run_id)
    #      mlflow.log_param("local_id", args.local_id)
    #      mlflow.log_param("args", args)
    #      mlflow.log_param("latest_dir", args.latest_dir)
    #      mlflow.log_param("local_dir", args.local_dir)
    #
    #      mlflow.log_param("experiment_params", experiment_params)
    #      experiment_params.log()


def archive_local_model(args, submission_file=None):
    #  if args.do_experiment:
    #      if submission_file:
    #          mlflow.log_param(f"{args.dataset_name}_submission_file",
    #                           submission_file)
    #          if os.path.exists(submission_file):
    #              mlflow.log_artifact(submission_file)
    #          logger.info(f"Log {submission_file} to tracking.mlflow.")

    if os.path.exists(args.local_dir):
        shutil.rmtree(args.local_dir)
    shutil.copytree(args.latest_dir, args.local_dir)
    logger.info(
        f"Archive local model({args.local_id}) {args.latest_dir} to {args.local_dir}"
    )

    #  os.remove(args.local_id_file)
    #  logger.info(
    #      f"Unlink {args.local_id_file} after archived local model({args.local_id})."
    #  )


#  def augement_entities(all_text_entities, labels_map):
#      aug_tokens = []
#      for i, (guid, text, entities) in enumerate(
#              tqdm(all_text_entities, desc=f"Augement {num_augements}X")):
#
#          #  print(f"-------------------{json_file}--------------------")
#          #  print(text)
#          #  print(entities)
#          #  for entity in entities:
#          #      s = entity['start_pos']
#          #      e = entity['end_pos']
#          #      print(f"{entity['label_type']}: {text[s:e]}")
#          #  print("----------------------------------------")
#          if entities:
#              for ai in range(num_augements):
#                  e_idx = random.randint(0, len(entities) - 1)
#                  entity = entities[e_idx]
#
#                  label_type = entity['label_type']
#                  s = entity['start_pos']
#                  e = entity['end_pos']
#
#                  labels = labels_map[label_type]
#                  idx = random.randint(0, len(labels) - 1)
#                  new_entity_text = labels[idx]
#
#                  text = text[:s] + new_entity_text + text[e:]
#
#                  assert len(new_entity_text) >= 0
#                  delta = len(new_entity_text) - (e - s)
#
#                  entity['end_pos'] = entity['start_pos'] + len(new_entity_text)
#                  entity['mention'] = new_entity_text
#
#                  assert text[
#                      entity['start_pos']:entity['end_pos']] == new_entity_text
#
#                  for n, e in enumerate(entities):
#                      if n > e_idx:
#                          e['start_pos'] += delta
#                          e['end_pos'] += delta
#
#                  aug_tokens.append(
#                      (f"{guid}-a{ai}", text, copy.deepcopy(entities)))
#
#      return aug_tokens


def tensor_to_numpy(t):
    return t.detach().cpu().numpy()


def tensor_to_list(t):
    return t.detach().cpu().numpy().to_list()


def save_args(args, model_path):
    if not os.path.exists(model_path):
        os.makedirs(model_path)
    #  torch.save(args, os.path.join(model_path, "training_args.bin"))
    logger.info(f"Save args in {model_path}/training_args.json")
    json.dump(
        {
            k: v
            for k, v in args.__dict__.items() if v is None
            or type(v) in [bool, str, int, float, dict, list, tuple]
        },
        open(os.path.join(model_path, "training_args.json"), 'w'),
        ensure_ascii=False,
        indent=2)



#  -*- coding: utf-8 -*-
#  File: __init__.py

# https://github.com/celery/kombu/blob/7d13f9b95d0b50c94393b962e6def928511bfda6/kombu/__init__.py#L34-L36
STATICA_HACK = True
globals()['kcah_acitats'[::-1].upper()] = False
if STATICA_HACK:
    from .base import *
    from .common import *
    from .format import *
    from .image import *
    from .parallel_map import *
    from .parallel import *
    from .raw import *
    from .remote import *
    from .serialize import *
    from . import imgaug
    from . import dataset


from pkgutil import iter_modules
import os
import os.path
from ..utils.develop import LazyLoader

__all__ = []


def _global_import(name):
    p = __import__(name, globals(), locals(), level=1)
    lst = p.__all__ if '__all__' in dir(p) else dir(p)
    if lst:
        del globals()[name]
        for k in lst:
            if not k.startswith('__'):
                globals()[k] = p.__dict__[k]
                __all__.append(k)


__SKIP = set(['dataset', 'imgaug'])
_CURR_DIR = os.path.dirname(__file__)
for _, module_name, __ in iter_modules(
        [os.path.dirname(__file__)]):
    srcpath = os.path.join(_CURR_DIR, module_name + '.py')
    if not os.path.isfile(srcpath):
        continue
    if "_test" not in module_name and \
       not module_name.startswith('_') and \
            module_name not in __SKIP:
        _global_import(module_name)


globals()['dataset'] = LazyLoader('dataset', globals(), __name__ + '.dataset')
globals()['imgaug'] = LazyLoader('imgaug', globals(), __name__ + '.imgaug')

del LazyLoader

__all__.extend(['imgaug', 'dataset'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import json
import os
import random
from copy import deepcopy
from datetime import datetime
from typing import Type, Union

import dill
import numpy as np
import pytorch_lightning as pl
import torch
import torch.functional as F
import torch.nn as nn
from loguru import logger
from pytorch_lightning.callbacks import EarlyStopping, ModelCheckpoint
from torch.nn import CrossEntropyLoss, MSELoss
from torch.optim.lr_scheduler import LambdaLR
from torch.utils.data import DataLoader
from tqdm import tqdm

from theta.nlp.arguments import (DataArguments, ModelArguments, TaskArguments,
                                 TrainingArguments,
                                 create_instance_from_arguments,
                                 generate_method_kwargs_from_arguments)
from transformers import (AdamW, AutoConfig, AutoModel, AutoTokenizer,
                          BertForSequenceClassification, ElectraConfig,
                          XLNetConfig)
from transformers.optimization import (
    get_cosine_schedule_with_warmup,
    get_cosine_with_hard_restarts_schedule_with_warmup,
    get_linear_schedule_with_warmup)

from ...utils import seed_everything

os.environ['TOKENIZERS_PARALLELISM'] = "true"


# ------------------------------ Dataset ------------------------------
class BaseDataset(object):
    """
        Dataset负责完成传入模型数据的编码工作
        编码data_generator生成的数据
    """
    def __init__(self, data_args, data_generator, label2id, tokenizer):
        super(BaseDataset, self).__init__()
        self.data_args = data_args
        self.label2id = label2id
        self.tokenizer = tokenizer
        self.sids = []
        self.seg_spans = []

        self.encoded_data_list = []
        for x in tqdm(data_generator(), desc="Encoding"):
            encoded = self._encode_item(x)
            if isinstance(encoded, list):
                self.encoded_data_list.extend(encoded)
                sids = [f"{x[0]}-{i}" for i in range(len(encoded))]
                self.sids.extend(sids)
            else:
                self.encoded_data_list.append(encoded)
                self.sids.append(x[0])

    def _encode_item(self, x):
        raise NotImplementedError

    def __iter__(self):
        for x in self.encoded_data_list:
            yield x

    def __getitem__(self, idx):
        return self.encoded_data_list[idx]

    def __len__(self):
        return len(self.encoded_data_list)


# ------------------------------ TaskData ------------------------------
class TaskData:
    """
    TaskData负责完成Samples -> Dataset的转换
    对模型提供train_dataset、val_dataset、test_dataset
    """
    def __init__(self,
                 data_args: Type[DataArguments],
                 train_samples,
                 test_samples,
                 val_samples=None,
                 tokenizer=None):
        self.data_args = data_args
        self.train_samples = train_samples
        self.test_samples = test_samples
        self.val_samples = val_samples
        self.tokenizer = tokenizer

        self._train_dataset = None
        self._val_dataset = None
        self._test_dataset = None

        if train_samples is not None and test_samples is not None:
            assert train_samples.label2id == test_samples.label2id
        if train_samples is not None:
            self.label2id = train_samples.label2id
        elif test_samples is not None:
            self.label2id = test_samples.label2id
        else:
            raise ValueError(
                f"Either train_samples or test_samples must be not None.")

    def load_train_data(self):
        self.train_samples.load_samples()

        if self.val_samples is not None:
            selv.val_samples.load_samples()
            self._splitted_train_samples = self.train_samples
            self._splitted_val_samples = self.val_samples
        else:
            all_train_samples = self.train_samples.shuffle()
            self._splitted_train_samples, self._splitted_val_samples = all_train_samples.split(
                ratios=self.data_args.split_ratios,
                random_state=self.data_args.random_state)
            logger.info(f"total train samples: {len(self.train_samples)}")
        logger.info(f"train samples: {len(self._splitted_train_samples)}")
        logger.info(f"val samples: {len(self._splitted_val_samples)}")

    def load_test_data(self):
        self.test_samples.load_samples()
        logger.info(f"test samples: {len(self.test_samples)}")

    def cache_dataset(self, mode, build_dataset_fn):
        cache_file = f"{self.data_args.cache_dir}/{mode}_dataset.cache"
        if not self.data_args.overwrite_cache and os.path.exists(cache_file):
            logger.info(f"Load cached {mode} dataset from {cache_file}")
            dataset = dill.load(open(cache_file, 'rb'))
            logger.warning(f"{len(dataset)} lines loaded.")
        else:
            dataset = build_dataset_fn()
            dill.dump(dataset, open(cache_file, 'wb'))
            logger.info(
                f"Save cached {mode} dataset {len(dataset)} lines to {cache_file}"
            )
        return dataset

    @property
    def train_dataset(self):
        if self._train_dataset is None:
            self._train_dataset = self.cache_dataset("train",
                                                     self.build_train_dataset)
        return self._train_dataset

    @property
    def val_dataset(self):
        if self._val_dataset is None:
            self._val_dataset = self.cache_dataset("val",
                                                   self.build_val_dataset)
        return self._val_dataset

    @property
    def test_dataset(self):
        if self._test_dataset is None:
            self._test_dataset = self.cache_dataset("test",
                                                    self.build_test_dataset)
        return self._test_dataset


def save_args(args, model_path):
    os.makedirs(model_path, exist_ok=True)
    args_file = os.path.join(model_path, "training_args.json")
    logger.info(f"Save args in {args_file}")
    json.dump(
        {
            k: v
            for k, v in args.__dict__.items() if v is None
            or type(v) in [bool, str, int, float, dict, list, tuple]
        },
        open(args_file, 'w'),
        ensure_ascii=False,
        indent=2)


class TransformerModel(nn.Module):
    def __init__(self,
                 model_name_or_path,
                 tokenizer=None,
                 automodel_cls=AutoModel):
        super(TransformerModel, self).__init__()
        assert automodel_cls is not None
        self.automodel_cls = automodel_cls
        logger.info(f"automodel_cls: {self.automodel_cls}")
        self.config = AutoConfig.from_pretrained(model_name_or_path)
        logger.info(f"{self.config}")

        if tokenizer is None:
            #  if model_name_or_path not in ['clue/roberta_chinese_pair_large']:
            #      tokenizer = AutoTokenizer.from_pretrained(model_name_or_path,
            #                                                fast=True)
            #  else:
            #      # self.config.vocab_size == 21128
            #      tokenizer = AutoTokenizer.from_pretrained("bert-base-chinese",
            #                                                fast=True)
            tokenizer = AutoTokenizer.from_pretrained("bert-base-chinese",
                                                      fast=True)
        self.tokenizer = tokenizer

        self.model_name_or_path = model_name_or_path
        #  self.transformer = self.automodel_cls.from_config(self.config)
        self.load_from_config()

    def init_weights(self):
        self.apply(self._init_weights)

    def _init_weights(self, module):
        """
        Initialize the weights
        Derived from BertPreTrainedModel._init_weights() in modeling_bert.py of transformers.
        """
        if isinstance(module, nn.Linear):
            # Slightly different from the TF version which uses truncated_normal for initialization
            # cf https://github.com/pytorch/pytorch/pull/5617
            module.weight.data.normal_(mean=0.0,
                                       std=self.config.initializer_range)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0,
                                       std=self.config.initializer_range)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    #  @staticmethod
    #  def _init_weights(blocks, **kwargs):
    #      """
    #      参数初始化，将 Linear / Embedding / LayerNorm 与 Bert 进行一样的初始化
    #      """
    #      for block in blocks:
    #          for module in block.modules():
    #              if isinstance(module, nn.Linear):
    #                  if module.bias is not None:
    #                      nn.init.zeros_(module.bias)
    #              elif isinstance(module, nn.Embedding):
    #                  nn.init.normal_(module.weight,
    #                                  mean=0,
    #                                  std=kwargs.pop('initializer_range', 0.02))
    #              elif isinstance(module, nn.LayerNorm):
    #                  nn.init.ones_(module.weight)
    #                  nn.init.zeros_(module.bias)

    def _is_xlnet(self):
        if isinstance(self.config, XLNetConfig):
            return True
        else:
            return False

    #  def load_from_pretrained(self, model_path):
    #      assert self.automodel_cls is not None
    #      self.transformer = self.automodel_cls.from_pretrained(model_path)
    #
    #      device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    #      self.transformer.to(device)

    #  def _load_from_pretrained(self, model_path, automodel_cls=None):
    #      if automodel_cls is not None:
    #          self.transformer = automodel_cls.from_pretrained(model_path)
    #      else:
    #          self.transformer = AutoModel.from_pretrained(model_path)
    #          #  self.transformer = BertForSequenceClassification.from_pretrained(model_path)
    #
    #      device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    #      self.transformer.to(device)

    def save_model(self, model_path):
        os.makedirs(model_path, exist_ok=True)
        #  model_to_save = (self.transformer.module if hasattr(
        #      self.transformer, "module") else self.transformer)
        #  model_to_save.save_pretrained(model_path)

        self.config.save_pretrained(os.path.abspath(model_path) + '/')
        self.tokenizer.save_vocabulary(os.path.abspath(model_path) + '/')

    #  def train(self):
    #      self.transformer.train()
    #
    #  def eval(self):
    #      self.transformer.eval()
    #
    #  def __call__(self, *args, **kwargs):
    #      outputs = self.transformer(*args, **kwargs)
    #      return outputs

    def forward(self, *args, **kwargs):
        outputs = self.transformer(*args, **kwargs)
        return outputs


# ------------------------------ TaskRunner ------------------------------
class TaskRunner(pl.LightningModule):
    def __init__(self, *args, **kwargs):
        super(TaskRunner, self).__init__()
        self.save_hyperparameters()

        #  model_name_or_path = self.hparams.model_name_or_path
        #  self.tokenizer = tokenizer

        #  if model_name_or_path is None:
        #      model_name_or_path = os.path.join(self.hparams.task_dir,
        #                                        "checkpoint")
        #  self.transformer_model = TransformerModel(model_name_or_path,
        #                                            tokenizer)

        self.warmup_steps = None
        self.total_steps = None
        self.wait_count = 0

        self.best_score = 0.0 if self.hparams.greater_is_better else float(
            'inf')

    #  @property
    #  def tokenizer(self):
    #      return self.transformer_model.tokenizer

    def _is_xlnet(self):
        config = self.transformer_model.config
        if isinstance(config, XLNetConfig):
            return True
        else:
            return False
        #  return 'XLNetLMHeadModel' in config.architectures
    def _is_electra(self):
        if isinstance(self.transformer_model.config, ElectraConfig):
            return True
        else:
            return False

    #  def load_model(self, model_path):
    #      pl_model_checkpoint_file = f"{model_path}/pl_model.ckpt"
    #      if os.path.exists(pl_model_checkpoint_file):
    #          logger.info(f"Load PL module from {pl_model_checkpoint_file}.")
    #          checkpoint = dill.load(open(pl_model_checkpoint_file, 'rb'))
    #          self.load_state_dict(checkpoint['state_dict'])
    #      else:
    #          #  self.model.load_model(model_path)
    #          logger.info(f"Load transformer model from {model_path}")
    #          self.model.load_model(model_path)

    def load_from_checkpoint(self, model_path):
        #  self.model.load_from_config()
        if os.path.isdir(model_path):
            pl_model_checkpoint_file = f"{model_path}/pl_model.ckpt"
        else:
            pl_model_checkpoint_file = model_path
        if os.path.exists(pl_model_checkpoint_file):
            logger.info(f"Load PL module from {pl_model_checkpoint_file}.")
            checkpoint = dill.load(open(pl_model_checkpoint_file, 'rb'))
            self.load_state_dict(checkpoint['state_dict'])
        else:
            logger.error(
                f"Checkpoint file {pl_model_checkpoint_file} does not exists.")

    def load_from_pretrained(self, model_name_or_path):
        logger.info(f"Load from pretrained model by {model_name_or_path}")
        self.model.load_from_pretrained(model_name_or_path)

    def save_model(self, model_path):
        os.makedirs(model_path, exist_ok=True)

        pl_model_checkpoint_file = f"{model_path}/pl_model.ckpt"

        # FIXME
        dill.dump({'state_dict': self.state_dict()},
                  open(pl_model_checkpoint_file, 'wb'))
        logger.warning(f"Save PL module in {pl_model_checkpoint_file}")

        self.model.save_model(model_path)
        logger.warning(f"Save transformer model in {model_path}")

    def configure_optimizers(self):
        #  param_optimizer = list(self.transformer_model.model.named_parameters())
        param_optimizer = list(self.model.named_parameters())
        no_decay = ["bias", "gamma", "beta"]
        optimizer_grouped_parameters = [
            {
                "params": [
                    p for n, p in param_optimizer
                    if not any(nd in n for nd in no_decay)
                ],
                "weight_decay_rate":
                self.hparams.weight_decay
            },
            {
                "params": [
                    p for n, p in param_optimizer
                    if any(nd in n for nd in no_decay)
                ],
                "weight_decay_rate":
                0.0
            },
        ]
        optimizer = AdamW(optimizer_grouped_parameters,
                          lr=self.hparams.learning_rate)  #,
        #  correct_bias=False)

        #  return optimizer

        assert self.warmup_steps is not None
        assert self.total_steps is not None

        # Inspired from https://github.com/PyTorchLightning/pytorch-lightning/issues/328
        #  def lr_exp_warmup(steps):
        #      if steps < self.warmup_steps:
        #          lr_scale = 0.1**(self.warmup_steps - steps)
        #      else:
        #          lr_scale = 0.95**steps
        #
        #      return lr_scale
        #
        #  scheduler = LambdaLR(optimizer, lr_lambda=lr_exp_warmup)

        schedule_fn = get_linear_schedule_with_warmup
        #  schedule_fn = get_cosine_schedule_with_warmup
        #  schedule_fn = get_cosine_with_hard_restarts_schedule_with_warmup
        scheduler = schedule_fn(optimizer,
                                num_warmup_steps=self.warmup_steps,
                                num_training_steps=self.total_steps)

        scheduler_dict = {
            'scheduler': scheduler,
            'interval': 'step',
            'frequency': 1
        }

        return ([optimizer], [scheduler_dict])

    #  def training_step_end(self, batch_parts_outputs):
    #      if self.trainer.global_step > 0 and self.trainer.global_step % self.hparams.eval_steps == 0:
    #          self.trainer.run_evaluation()

    @property
    def current_epoch(self):
        return self.trainer.current_epoch

    @property
    def max_epochs(self):
        return self.trainer.max_epochs

    #  def on_train_start(self):
    #      self.model.train()
    #
    #  def on_validation_start(self, model):
    #      self.model.eval()
    #
    #  def on_validation_end(self, model):
    #      self.model.train()
    #
    #  def on_test_start(self, model):
    #      self.model.eval()
    #
    #  def on_test_end(self, model):
    #      self.model.train()

    def save_best_model(self, eval_outputs: dict):

        if 'val_loss' in eval_outputs:
            val_loss = eval_outputs['val_loss']
            logger.info(
                f"Epoch {self.current_epoch}/{self.max_epochs} val_loss: {val_loss:.4f}"
            )
        else:
            logger.warning(f"No val_loss in eval_outputs: {eval_outputs}")

        if self.hparams.metric_for_best_model in eval_outputs:
            curr_score = eval_outputs[self.hparams.metric_for_best_model]

            is_best = False
            if self.hparams.greater_is_better:
                if curr_score > self.best_score:
                    is_best = True
            else:
                if curr_score < self.best_score:
                    is_best = True

            if is_best:
                self.wait_count = 0
                logger.warning(
                    f"Best {self.hparams.metric_for_best_model}: {curr_score:.4f} / {self.best_score:.4f}"
                )
                self.best_score = curr_score

                self.save_model(
                    os.path.join(self.hparams.task_dir, "checkpoint"))
            else:
                self.wait_count += 1
                logger.info(
                    f"{self.hparams.metric_for_best_model}: {curr_score:.4f} / {self.best_score:.4f}, earlystopping_patience: {self.wait_count}/{self.hparams.earlystopping_patience}"
                )
                if self.wait_count >= self.hparams.earlystopping_patience:
                    self.trainer.should_stop = True


# ------------------------------ Task ------------------------------


class BaseTask():
    """
    实验任务(ExperimentTask)
    定义一次任务中必须包含的数据、模型、参数。
    支持任务执行、加载复现。

    task_args = TaskArguments.parse_args()
    task_data = ExpData(data_args=task_args.data_args)
    task_trainer = ExpModel(model_args=task_args.model_args,
                          num_labels=len(glue_labels))

    task = ExpTask(args=task_args,
                   data=task_data,
                   trainer=task_trainer)

    task.execute(glue_labels)

    """
    def __init__(self, args: Type[TaskArguments], data: Type[TaskData],
                 runner: Type[TaskRunner]):
        self.args = args
        self.data = data
        self.runner = runner

        os.environ['TOKENIZERS_PARALLELISM'] = "true"
        seed_everything(self.args.training_args.seed)  #固定随机种子

        #  self.args.training_args.device = torch.device(
        #      'cuda' if torch.cuda.is_available() else 'cpu')

    @property
    def rootdir(self):
        return self.args.rootdir

    @property
    def data_args(self):
        return self.args.data_args

    @property
    def model_args(self):
        return self.args.model_args

    @property
    def training_args(self):
        return self.args.training_args

    @property
    def remaining_args(self):
        return self.args.remaining_args

    @property
    def latest_path(self):
        return self.training_args.latest_path

    @property
    def checkpoint_path(self):
        return self.model_args.checkpoint_path

    @property
    def test_results_file(self):
        return os.path.join(self.training_args.task_dir, "test_results.pkl")

    def load_test_results(self, test_results_file=None):
        if test_results_file is None:
            test_results_file = os.path.join(self.latest_path,
                                             "test_results.pkl")
        logger.info(f"Load test result from {test_results_file}")
        test_results = dill.load(open(test_results_file, 'rb'))
        return test_results

    def dump_test_results(self, test_results, test_results_file=None):
        if test_results_file is None:
            test_results_file = self.test_results_file
        dill.dump(test_results, open(test_results_file, 'wb'))
        logger.info(f"Dump test results to {test_results_file}.")

        return test_results_file

    @property
    def train_dataloader(self):
        train_dataset = self.data.train_dataset
        for index in random.sample(range(len(train_dataset)), 1):
            logger.info(
                f"Sample {index} of the training set: {train_dataset[index]}.")

        train_dataloader = DataLoader(
            train_dataset,
            batch_size=self.training_args.per_device_train_batch_size,
            collate_fn=train_dataset.collate_fn,
            pin_memory=True,
            num_workers=8)
        return train_dataloader

    @property
    def val_dataloader(self):
        val_dataset = self.data.val_dataset
        for index in random.sample(range(len(val_dataset)), 1):
            logger.info(
                f"Sample {index} of the val set: {val_dataset[index]}.")

        val_dataloader = DataLoader(
            val_dataset,
            batch_size=self.training_args.per_device_eval_batch_size,
            collate_fn=val_dataset.collate_fn,
            pin_memory=True,
            num_workers=8)
        return val_dataloader

    @property
    def test_dataloader(self):
        test_dataset = self.data.test_dataset
        for index in random.sample(range(len(test_dataset)), 1):
            logger.info(
                f"Sample {index} of the test set: {test_dataset[index]}.")

        test_dataloader = DataLoader(
            test_dataset,
            batch_size=self.training_args.per_device_test_batch_size,
            collate_fn=test_dataset.collate_fn,
            pin_memory=True,
            num_workers=8)
        return test_dataloader

    def get_latest_submission_file(self, ext="json", prefix="submission"):
        ts = datetime.now().strftime("%Y%m%d.%H%M%S")
        real_path = os.readlink(self.latest_path)
        task_id = os.path.basename(real_path)
        submission_file = os.path.join(self.training_args.submissions_dir,
                                       f"{prefix}_{task_id}_{ts}.{ext}")
        return submission_file

    def generate_submission(self):
        logger.warning(
            f"Call generate_submission() implemented in base class Task.")

    def execute(self, *args, **kwargs):

        model_args = self.model_args
        data_args = self.data_args
        training_args = self.training_args
        remaining_args = self.remaining_args

        return_dict = {}

        # ------------------------------ do_train ------------------------------
        if training_args.do_train:
            model_path = model_args.model_name_or_path
            self.runner.load_from_pretrained(model_path)
            self.data.tokenizer = self.runner.model.tokenizer
            self.data.load_train_data()

            def setup_warmup_steps():
                epoch_steps = int(
                    len(self.data.train_dataset) /
                    training_args.per_device_train_batch_size)

                warmup_method = training_args.warmup_method
                if warmup_method == 'auto':
                    if training_args.max_steps is not None:
                        training_args.warmup_steps = int(
                            training_args.max_steps / 10)
                    else:
                        training_args.warmup_steps = int(
                            epoch_steps * training_args.max_epochs / 10)
                elif warmup_method == 'by_epoch':
                    training_args.warmup_steps = epoch_steps * training_args.warmup_epochs
                elif warmup_method == 'by_rate':
                    if training_args.max_steps is not None:
                        training_args.warmup_steps = int(
                            training_args.max_steps *
                            training_args.warmup_rate)
                    else:
                        training_args.warmup_steps = int(
                            epoch_steps * training_args.max_epochs *
                            training_args.warmup_rate)
                elif warmup_method == 'by_steps':
                    if training_args.warmup_steps > 0:
                        pass
                    else:
                        training_args.warmup_steps = epoch_steps * training_args.warmup_epochs

            setup_warmup_steps()

            max_epochs = self.training_args.max_epochs
            if max_epochs:
                total_steps = int(
                    len(self.data.train_dataset) / self.training_args.
                    per_device_train_batch_size) * max_epochs
            else:
                total_steps = self.training_args.max_steps

            self.runner.warmup_steps = training_args.warmup_steps
            self.runner.total_steps = total_steps

            #  self.training_args.max_steps = total_steps

            logger.warning(f"warmup_steps: {self.training_args.warmup_steps}")
            logger.warning(f"total_steps: {total_steps}")

            val_epoch_steps = int(
                len(self.data.val_dataset) /
                training_args.per_device_eval_batch_size)

            #  trainer = create_instance_from_arguments(
            #      pl.Trainer, self.training_args.to_dict())
            trainer_kwargs = generate_method_kwargs_from_arguments(
                pl.Trainer, method="__init__", args=training_args.to_dict())

            trainer_kwargs['precision'] = 16 if training_args.fp16 else 32

            if training_args.eval_steps_or_interval >= 10:
                trainer_kwargs[
                    'val_check_interval'] = training_args.eval_steps_or_interval / epoch_steps
            else:
                trainer_kwargs[
                    'val_check_interval'] = training_args.eval_steps_or_interval

            trainer_kwargs['checkpoint_callback'] = False
            trainer_kwargs['num_sanity_val_steps'] = int(val_epoch_steps * 0.5)

            def symlink_latest_path(latest_path):
                if os.path.islink(latest_path):
                    os.remove(latest_path)
                os.symlink(os.path.basename(training_args.task_dir),
                           latest_path)

            symlink_latest_path(self.latest_path)

            def save_trainer_kwargs(save_dir):
                trainer_kwargs_file = os.path.join(save_dir,
                                                   "trainer_kwargs.json")
                json.dump(trainer_kwargs,
                          open(trainer_kwargs_file, 'w'),
                          ensure_ascii=False,
                          indent=2)
                logger.info(f"Saved trainer_kwargs to {trainer_kwargs_file}")

            save_trainer_kwargs(self.latest_path)

            #  if training_args.earlystopping_patience > 0:
            #      earlystopping_cb = EarlyStopping(
            #          monitor=training_args.metric_for_best_model,
            #          patience=training_args.earlystopping_patience)
            #      trainer_kwargs['callbacks'] = [earlystopping_cb]
            #
            logger.warning(f"trainer_kwargs: {trainer_kwargs}")

            trainer = pl.Trainer(**trainer_kwargs)
            trainer.fit(self.runner, self.train_dataloader,
                        self.val_dataloader)

        # ------------------------------ do_eval ------------------------------
        if training_args.do_eval:
            self.runner.load_from_checkpoint(self.checkpoint_path)
            self.data.tokenizer = self.runner.model.tokenizer
            self.data.load_train_data()

            val_dataloader = self.val_dataloader
            #  eval_results = do_eval(trainer, val_dataset, data_args)

        # ------------------------------ do_predict ------------------------------
        if training_args.do_predict:
            self.runner.load_from_checkpoint(self.checkpoint_path)
            self.data.tokenizer = self.runner.model.tokenizer
            self.data.load_test_data()

            test_dataloader = self.test_dataloader

            trainer_kwargs = generate_method_kwargs_from_arguments(
                pl.Trainer,
                method="__init__",
                args=self.training_args.to_dict())

            trainer_kwargs['precision'] = 16 if self.training_args.fp16 else 32

            trainer = pl.Trainer(**trainer_kwargs)

            trainer.test(self.runner, test_dataloaders=test_dataloader)

            test_results_file = self.dump_test_results(
                self.runner.test_results)

            return_dict['test_results_file'] = test_results_file

        if training_args.do_submit:
            submission_file = self.generate_submission()

            return_dict['submission_file'] = submission_file

        return return_dict



# -*- coding: utf-8 -*-
# File: utils.py


import inspect
import numpy as np
import re
import os
import sys
from contextlib import contextmanager
from datetime import datetime, timedelta
from tqdm import tqdm

from . import logger
from .concurrency import subproc_call

__all__ = ['change_env',
           'get_rng',
           'fix_rng_seed',
           'get_tqdm',
           'execute_only_once',
           'humanize_time_delta'
           ]


def humanize_time_delta(sec):
    """Humanize timedelta given in seconds

    Args:
        sec (float): time difference in seconds. Must be positive.

    Returns:
        str - time difference as a readable string

    Example:

    .. code-block:: python

        print(humanize_time_delta(1))                                   # 1 second
        print(humanize_time_delta(60 + 1))                              # 1 minute 1 second
        print(humanize_time_delta(87.6))                                # 1 minute 27 seconds
        print(humanize_time_delta(0.01))                                # 0.01 seconds
        print(humanize_time_delta(60 * 60 + 1))                         # 1 hour 1 second
        print(humanize_time_delta(60 * 60 * 24 + 1))                    # 1 day 1 second
        print(humanize_time_delta(60 * 60 * 24 + 60 * 2 + 60*60*9 + 3)) # 1 day 9 hours 2 minutes 3 seconds
    """
    if sec < 0:
        logger.warn("humanize_time_delta() obtains negative seconds!")
        return "{:.3g} seconds".format(sec)
    if sec == 0:
        return "0 second"
    time = datetime(2000, 1, 1) + timedelta(seconds=int(sec))
    units = ['day', 'hour', 'minute', 'second']
    vals = [int(sec // 86400), time.hour, time.minute, time.second]
    if sec < 60:
        vals[-1] = sec

    def _format(v, u):
        return "{:.3g} {}{}".format(v, u, "s" if v > 1 else "")

    ans = []
    for v, u in zip(vals, units):
        if v > 0:
            ans.append(_format(v, u))
    return " ".join(ans)


@contextmanager
def change_env(name, val):
    """
    Args:
        name(str): name of the env var
        val(str or None): the value, or set to None to clear the env var.

    Returns:
        a context where the environment variable ``name`` being set to
        ``val``. It will be set back after the context exits.
    """
    oldval = os.environ.get(name, None)

    if val is None:
        try:
            del os.environ[name]
        except KeyError:
            pass
    else:
        os.environ[name] = val

    yield

    if oldval is None:
        try:
            del os.environ[name]
        except KeyError:
            pass
    else:
        os.environ[name] = oldval


_RNG_SEED = None


def fix_rng_seed(seed):
    """
    Call this function at the beginning of program to fix rng seed within tensorpack.

    Args:
        seed (int):

    Note:
        See https://github.com/tensorpack/tensorpack/issues/196.

    Example:

        Fix random seed in both tensorpack and tensorflow.

    .. code-block:: python

            seed = 42
            utils.fix_rng_seed(seed)
            tesnorflow.set_random_seed(seed)
            # run trainer
    """
    global _RNG_SEED
    _RNG_SEED = int(seed)


def get_rng(obj=None):
    """
    Get a good RNG seeded with time, pid and the object.

    Args:
        obj: some object to use to generate random seed.
    Returns:
        np.random.RandomState: the RNG.
    """
    seed = (id(obj) + os.getpid() +
            int(datetime.now().strftime("%Y%m%d%H%M%S%f"))) % 4294967295
    if _RNG_SEED is not None:
        seed = _RNG_SEED
    return np.random.RandomState(seed)


_EXECUTE_HISTORY = set()


def execute_only_once():
    """
    Each called in the code to this function is guaranteed to return True the
    first time and False afterwards.

    Returns:
        bool: whether this is the first time this function gets called from this line of code.

    Example:
        .. code-block:: python

            if execute_only_once():
                # do something only once
    """
    f = inspect.currentframe().f_back
    ident = (f.f_code.co_filename, f.f_lineno)
    if ident in _EXECUTE_HISTORY:
        return False
    _EXECUTE_HISTORY.add(ident)
    return True


def _pick_tqdm_interval(file):
    # Heuristics to pick a update interval for progress bar that's nice-looking for users.
    isatty = file.isatty()
    # Jupyter notebook should be recognized as tty.
    # Wait for https://github.com/ipython/ipykernel/issues/268
    try:
        from ipykernel import iostream
        if isinstance(file, iostream.OutStream):
            isatty = True
    except ImportError:
        pass

    if isatty:
        return 0.5
    else:
        # When run under mpirun/slurm, isatty is always False.
        # Here we apply some hacky heuristics for slurm.
        if 'SLURM_JOB_ID' in os.environ:
            if int(os.environ.get('SLURM_JOB_NUM_NODES', 1)) > 1:
                # multi-machine job, probably not interactive
                return 60
            else:
                # possibly interactive, so let's be conservative
                return 15

        if 'OMPI_COMM_WORLD_SIZE' in os.environ:
            return 60

        # If not a tty, don't refresh progress bar that often
        return 180


def get_tqdm_kwargs(**kwargs):
    """
    Return default arguments to be used with tqdm.

    Args:
        kwargs: extra arguments to be used.
    Returns:
        dict:
    """
    default = dict(
        smoothing=0.5,
        dynamic_ncols=True,
        ascii=True,
        bar_format='{l_bar}{bar}|{n_fmt}/{total_fmt}[{elapsed}<{remaining},{rate_noinv_fmt}]'
    )

    try:
        # Use this env var to override the refresh interval setting
        interval = float(os.environ['TENSORPACK_PROGRESS_REFRESH'])
    except KeyError:
        interval = _pick_tqdm_interval(kwargs.get('file', sys.stderr))

    default['mininterval'] = interval
    default.update(kwargs)
    return default


def get_tqdm(*args, **kwargs):
    """ Similar to :func:`tqdm.tqdm()`,
    but use tensorpack's default options to have consistent style. """
    return tqdm(*args, **get_tqdm_kwargs(**kwargs))


def find_library_full_path(name):
    """
    Similar to `from ctypes.util import find_library`, but try
    to return full path if possible.
    """
    from ctypes.util import find_library

    if os.name == "posix" and sys.platform == "darwin":
        # on Mac, ctypes already returns full path
        return find_library(name)

    def _use_proc_maps(name):
        """
        Find so from /proc/pid/maps
        Only works with libraries that has already been loaded.
        But this is the most accurate method -- it finds the exact library that's being used.
        """
        procmap = os.path.join('/proc', str(os.getpid()), 'maps')
        if not os.path.isfile(procmap):
            return None
        try:
            with open(procmap, 'r') as f:
                for line in f:
                    line = line.strip().split(' ')
                    sofile = line[-1]

                    basename = os.path.basename(sofile)
                    if 'lib' + name + '.so' in basename:
                        if os.path.isfile(sofile):
                            return os.path.realpath(sofile)
        except IOError:
            # can fail in certain environment (e.g. chroot)
            # if the pids are incorrectly mapped
            pass

    # The following two methods come from https://github.com/python/cpython/blob/master/Lib/ctypes/util.py
    def _use_ld(name):
        """
        Find so with `ld -lname -Lpath`.
        It will search for files in LD_LIBRARY_PATH, but not in ldconfig.
        """
        cmd = "ld -t -l{} -o {}".format(name, os.devnull)
        ld_lib_path = os.environ.get('LD_LIBRARY_PATH', '')
        for d in ld_lib_path.split(':'):
            cmd = cmd + " -L " + d
        result, ret = subproc_call(cmd + '|| true')
        expr = r'[^\(\)\s]*lib%s\.[^\(\)\s]*' % re.escape(name)
        res = re.search(expr, result.decode('utf-8'))
        if res:
            res = res.group(0)
            if not os.path.isfile(res):
                return None
            return os.path.realpath(res)

    def _use_ldconfig(name):
        """
        Find so in `ldconfig -p`.
        It does not handle LD_LIBRARY_PATH.
        """
        with change_env('LC_ALL', 'C'), change_env('LANG', 'C'):
            ldconfig, ret = subproc_call("ldconfig -p")
            ldconfig = ldconfig.decode('utf-8')
            if ret != 0:
                return None
        expr = r'\s+(lib%s\.[^\s]+)\s+\(.*=>\s+(.*)' % (re.escape(name))
        res = re.search(expr, ldconfig)
        if not res:
            return None
        else:
            ret = res.group(2)
            return os.path.realpath(ret)

    if sys.platform.startswith('linux'):
        return _use_proc_maps(name) or _use_ld(name) or _use_ldconfig(name) or find_library(name)

    return find_library(name)  # don't know what to do



# -*- coding: utf-8 -*-
# File: convert.py

import numpy as np
import cv2

from .base import PhotometricAugmentor

__all__ = ['ColorSpace', 'Grayscale', 'ToUint8', 'ToFloat32']


class ColorSpace(PhotometricAugmentor):
    """ Convert into another color space.  """

    def __init__(self, mode, keepdims=True):
        """
        Args:
            mode: OpenCV color space conversion code (e.g., ``cv2.COLOR_BGR2HSV``)
            keepdims (bool): keep the dimension of image unchanged if OpenCV
                changes it.
        """
        super(ColorSpace, self).__init__()
        self._init(locals())

    def _augment(self, img, _):
        transf = cv2.cvtColor(img, self.mode)
        if self.keepdims:
            if len(transf.shape) is not len(img.shape):
                transf = transf[..., None]
        return transf


class Grayscale(ColorSpace):
    """ Convert RGB or BGR image to grayscale. """

    def __init__(self, keepdims=True, rgb=False, keepshape=False):
        """
        Args:
            keepdims (bool): return image of shape [H, W, 1] instead of [H, W]
            rgb (bool): interpret input as RGB instead of the default BGR
            keepshape (bool): whether to duplicate the gray image into 3 channels
                so the result has the same shape as input.
        """
        mode = cv2.COLOR_RGB2GRAY if rgb else cv2.COLOR_BGR2GRAY
        if keepshape:
            assert keepdims, "keepdims must be True when keepshape==True"
        super(Grayscale, self).__init__(mode, keepdims)
        self.keepshape = keepshape
        self.rgb = rgb

    def _augment(self, img, _):
        ret = super()._augment(img, _)
        if self.keepshape:
            return np.concatenate([ret] * 3, axis=2)
        else:
            return ret


class ToUint8(PhotometricAugmentor):
    """ Clip and convert image to uint8. Useful to reduce communication overhead. """
    def _augment(self, img, _):
        return np.clip(img, 0, 255).astype(np.uint8)


class ToFloat32(PhotometricAugmentor):
    """ Convert image to float32, may increase quality of the augmentor. """
    def _augment(self, img, _):
        return img.astype(np.float32)



# -*- coding: utf-8 -*-
# File: transform.py

import numpy as np
import cv2

from ...utils.argtools import log_once

from .base import ImageAugmentor, _default_repr

TransformAugmentorBase = ImageAugmentor
"""
Legacy alias. Please don't use.
"""
# This legacy augmentor requires us to import base from here, causing circular dependency.
# Should remove this in the future.

__all__ = ["Transform", "ResizeTransform", "CropTransform", "FlipTransform",
           "TransformList", "TransformFactory"]


# class WrappedImgFunc(object):
#     def __init__(self, func, need_float=False, cast_back=True, fix_ndim=True):
#         self.func = func
#         self.need_float = need_float
#         self.cast_back = cast_back

#     def __call__(self, img):
#         old_dtype = img.dtype
#         old_ndim = img.ndim

#         if self.need_float:
#             img = img.astype("float32")

#         img = self.func(img)

#         if self.cast_back and old_dtype == np.uint8 and img.dtype != np.uint8:
#             img = np.clip(img, 0, 255.)

#         if self.cast_back:
#             img = img.astype(old_dtype)

#         if self.fix_ndim and old_ndim == 3 and img.ndim == 2:
#             img = img[:, :, np.newaxis]
#         return img


class BaseTransform(object):
    """
    Base class for all transforms, for type-check only.

    Users should never interact with this class.
    """
    def _init(self, params=None):
        if params:
            for k, v in params.items():
                if k != 'self' and not k.startswith('_'):
                    setattr(self, k, v)


class Transform(BaseTransform):
    """
    A deterministic image transformation, used to implement
    the (probably random) augmentors.

    This class is also the place to provide a default implementation to any
    :meth:`apply_xxx` method.
    The current default is to raise NotImplementedError in any such methods.

    All subclasses should implement `apply_image`.
    The image should be of type uint8 in range [0, 255], or
    floating point images in range [0, 1] or [0, 255]

    Some subclasses may implement `apply_coords`, when applicable.
    It should take and return a numpy array of Nx2, where each row is the (x, y) coordinate.

    The implementation of each method may choose to modify its input data
    in-place for efficient transformation.
    """

    def __init__(self):
        # provide an empty __init__, so that __repr__ will work nicely
        pass

    def __getattr__(self, name):
        if name.startswith("apply_"):

            def f(x):
                raise NotImplementedError("{} does not implement method {}".format(self.__class__.__name__, name))

            return f
        raise AttributeError("Transform object has no attribute {}".format(name))

    def __repr__(self):
        try:
            return _default_repr(self)
        except AssertionError as e:
            log_once(e.args[0], 'warn')
            return super(Transform, self).__repr__()

    __str__ = __repr__


class ResizeTransform(Transform):
    """
    Resize the image.
    """
    def __init__(self, h, w, new_h, new_w, interp):
        """
        Args:
            h, w (int):
            new_h, new_w (int):
            interp (int): cv2 interpolation method
        """
        super(ResizeTransform, self).__init__()
        self._init(locals())

    def apply_image(self, img):
        assert img.shape[:2] == (self.h, self.w)
        ret = cv2.resize(
            img, (self.new_w, self.new_h),
            interpolation=self.interp)
        if img.ndim == 3 and ret.ndim == 2:
            ret = ret[:, :, np.newaxis]
        return ret

    def apply_coords(self, coords):
        coords[:, 0] = coords[:, 0] * (self.new_w * 1.0 / self.w)
        coords[:, 1] = coords[:, 1] * (self.new_h * 1.0 / self.h)
        return coords


class CropTransform(Transform):
    """
    Crop a subimage from an image.
    """
    def __init__(self, y0, x0, h, w):
        super(CropTransform, self).__init__()
        self._init(locals())

    def apply_image(self, img):
        return img[self.y0:self.y0 + self.h, self.x0:self.x0 + self.w]

    def apply_coords(self, coords):
        coords[:, 0] -= self.x0
        coords[:, 1] -= self.y0
        return coords


class WarpAffineTransform(Transform):
    def __init__(self, mat, dsize, interp=cv2.INTER_LINEAR,
                 borderMode=cv2.BORDER_CONSTANT, borderValue=0):
        super(WarpAffineTransform, self).__init__()
        self._init(locals())

    def apply_image(self, img):
        ret = cv2.warpAffine(img, self.mat, self.dsize,
                             flags=self.interp,
                             borderMode=self.borderMode,
                             borderValue=self.borderValue)
        if img.ndim == 3 and ret.ndim == 2:
            ret = ret[:, :, np.newaxis]
        return ret

    def apply_coords(self, coords):
        coords = np.concatenate((coords, np.ones((coords.shape[0], 1), dtype='f4')), axis=1)
        coords = np.dot(coords, self.mat.T)
        return coords


class FlipTransform(Transform):
    """
    Flip the image.
    """
    def __init__(self, h, w, horiz=True):
        """
        Args:
            h, w (int):
            horiz (bool): whether to flip horizontally or vertically.
        """
        self._init(locals())

    def apply_image(self, img):
        if self.horiz:
            return img[:, ::-1]
        else:
            return img[::-1]

    def apply_coords(self, coords):
        if self.horiz:
            coords[:, 0] = self.w - coords[:, 0]
        else:
            coords[:, 1] = self.h - coords[:, 1]
        return coords


class TransposeTransform(Transform):
    """
    Transpose the image.
    """
    def apply_image(self, img):
        ret = cv2.transpose(img)
        if img.ndim == 3 and ret.ndim == 2:
            ret = ret[:, :, np.newaxis]
        return ret

    def apply_coords(self, coords):
        return coords[:, ::-1]


class NoOpTransform(Transform):
    """
    A Transform that does nothing.
    """
    def __getattr__(self, name):
        if name.startswith("apply_"):
            return lambda x: x
        raise AttributeError("NoOpTransform object has no attribute {}".format(name))


class PhotometricTransform(Transform):
    """
    A transform which only has `apply_image` but does nothing in `apply_coords`.
    """
    def __init__(self, func, name=None):
        """
        Args:
            func (img -> img): a function to be used for :meth:`apply_image`
            name (str, optional): the name of this transform
        """
        self._func = func
        self._name = name

    def apply_image(self, img):
        return self._func(img)

    def apply_coords(self, coords):
        return coords

    def __repr__(self):
        return "imgaug.PhotometricTransform({})".format(self._name if self._name else "")

    __str__ = __repr__


class TransformFactory(Transform):
    """
    Create a :class:`Transform` from user-provided functions.
    """
    def __init__(self, name=None, **kwargs):
        """
        Args:
            name (str, optional): the name of this transform
            **kwargs: mapping from `'apply_xxx'` to implementation of such functions.
        """
        for k, v in kwargs.items():
            if k.startswith('apply_'):
                setattr(self, k, v)
            else:
                raise KeyError("Unknown argument '{}' in TransformFactory!".format(k))
        self._name = name

    def __str__(self):
        return "imgaug.TransformFactory({})".format(self._name if self._name else "")

    __repr__ = __str__


"""
Some meta-transforms:
they do not perform actual transformation, but delegate to another Transform.
"""


class TransformList(BaseTransform):
    """
    Apply a list of transforms sequentially.
    """
    def __init__(self, tfms):
        """
        Args:
            tfms (list[Transform]):
        """
        for t in tfms:
            assert isinstance(t, BaseTransform), t
        self.tfms = tfms

    def _apply(self, x, meth):
        for t in self.tfms:
            x = getattr(t, meth)(x)
        return x

    def __getattr__(self, name):
        if name.startswith("apply_"):
            return lambda x: self._apply(x, name)
        raise AttributeError("TransformList object has no attribute {}".format(name))

    def __str__(self):
        repr_each_tfm = ",\n".join(["  " + repr(x) for x in self.tfms])
        return "imgaug.TransformList([\n{}])".format(repr_each_tfm)

    def __add__(self, other):
        other = other.tfms if isinstance(other, TransformList) else [other]
        return TransformList(self.tfms + other)

    def __iadd__(self, other):
        other = other.tfms if isinstance(other, TransformList) else [other]
        self.tfms.extend(other)
        return self

    def __radd__(self, other):
        other = other.tfms if isinstance(other, TransformList) else [other]
        return TransformList(other + self.tfms)

    __repr__ = __str__


class LazyTransform(BaseTransform):
    """
    A transform that's instantiated at the first call to `apply_image`.
    """
    def __init__(self, get_transform):
        """
        Args:
            get_transform (img -> Transform): a function which will be used to instantiate a Transform.
        """
        self.get_transform = get_transform
        self._transform = None

    def apply_image(self, img):
        if not self._transform:
            self._transform = self.get_transform(img)
        return self._transform.apply_image(img)

    def _apply(self, x, meth):
        assert self._transform is not None, \
            "LazyTransform.{} can only be called after the transform has been applied on an image!"
        return getattr(self._transform, meth)(x)

    def __getattr__(self, name):
        if name.startswith("apply_"):
            return lambda x: self._apply(x, name)
        raise AttributeError("TransformList object has no attribute {}".format(name))

    def __repr__(self):
        if self._transform is None:
            return "LazyTransform(get_transform={})".format(str(self.get_transform))
        else:
            return repr(self._transform)

    __str__ = __repr__

    def apply_coords(self, coords):
        return self._apply(coords, "apply_coords")


if __name__ == '__main__':
    shape = (100, 100)
    center = (10, 70)
    mat = cv2.getRotationMatrix2D(center, 20, 1)
    trans = WarpAffineTransform(mat, (130, 130))

    def draw_points(img, pts):
        for p in pts:
            try:
                img[int(p[1]), int(p[0])] = 0
            except IndexError:
                pass

    image = cv2.imread('cat.jpg')
    image = cv2.resize(image, shape)
    orig_image = image.copy()
    coords = np.random.randint(100, size=(20, 2))

    draw_points(orig_image, coords)
    print(coords)

    for _ in range(1):
        coords = trans.apply_coords(coords)
        image = trans.apply_image(image)
    print(coords)
    draw_points(image, coords)

    # viz = cv2.resize(viz, (1200, 600))
    orig_image = cv2.resize(orig_image, (600, 600))
    image = cv2.resize(image, (600, 600))
    viz = np.concatenate((orig_image, image), axis=1)
    cv2.imshow("mat", viz)
    cv2.waitKey()
